#!/usr/bin/python
# -*- coding=utf-8 -*-
"""
Arabic numbers routins
@author: Taha Zerrouki
@contact: taha dot zerrouki at gmail dot com
@copyright: Arabtechies, Arabeyes, Taha Zerrouki
@license: GPL
@date:2017/02/14
@version: 0.3
# ArNumbers is imported from
license:   LGPL <http://www.gnu.org/licenses/lgpl.txt>
link      http://www.ar-php.org
category  Text
author    Khaled Al-Shamaa <khaled.alshamaa@gmail.com>
copyright 2009 Khaled Al-Shamaa
"""
import math
import sys

from . import arabrepr, araby
from . import named_const as nmconst
from . import number_const as nbconst


class ArNumbers(object):
    """
    Arabic number class
    """

    _individual = {}
    _feminine = 1
    _format = 1

    ##"""
    ##     * Loads initialize values
    ##"""
    def __init__(self):
        self._individual = nbconst.INDIVIDUALS
        self.complications = nbconst.COMPLICATIONS

    def set_feminine(self, value):
        """
        Set feminine flag of the counted object
        @param value: value Counted object feminine (1 for masculine & 2 for feminine)
        @type value: integer
        @return: True if success, or False if fail
        @rtype: boolean
        """

        flag = True
        if value in (1, 2):
            self._feminine = value
        else:
            flag = False
        return flag

    def set_format(self, value):
        """
        Set the grammar position flag of the counted object
        @param value: Grammar position of counted object (1 if Marfoua & 2 if Mansoub or Majrour)
        @type value: integer
        @return: True if success, or False if fail
        @rtype: boolean
        """

        flag = True

        if value in (1, 2):
            self._format = value
        else:
            flag = False
        return flag

    def get_feminine(self):
        """
        Get the feminine flag of counted object
        @return: return current setting of counted object feminine flag
        @rtype: integer
        """
        return self._feminine

    def get_format(self):
        """
        Get the grammer position flag of counted object
        @return: return current setting of counted object grammer position flag
        @rtype: integer
        """

        return self._format

    def int2str(self, number, output_charset=None, main=None):
        """
        Spell integer number in Arabic idiom
        @param number: The number you want to spell in Arabic idiom
        @type number: integer
        @param  output_charset: (optional) Output charset [utf-8|windows-1256|iso-8859-6]
        default value is None (use set output charset)
        @type  output_charset: string
        @param main:  Main Ar-PHP object to access charset converter options
        @type main: object
        @return: The Arabic idiom that spells inserted number
        @rtype: string
        """
        temp = number.split(".")
        string = self._int2str(temp[0])
        if len(temp) > 1:
            dec = self._int2str(temp[1])
            string += " فاصلة " + dec
        if main:
            if output_charset is None:
                output_charset = main.getOutputCharset()
            string = main.coreConvert(string, "utf-8", output_charset)
        return string

    def _int2str(self, number_str):
        """
        Spell integer number in Arabic idiom
        @param number_str: The number you want to spell in Arabic idiom
        @type number_str: integer.
        @return: The Arabic idiom that spells inserted number
        @rtype:string
        """

        blocks = []
        items = []
        string = ""
        # ~ number = number#trunc(int(number)) #(int)number)
        try:
            number = int(number_str)
        except ValueError:
            number = 0
        if int(number) > 0:
            number_part = ""
            while len(number_str) > 3:
                blocks.append(number_str[-3:])
                number_str = number_str[: len(number_str) - 3]
            blocks.append(number_str)
            blocks_num = len(blocks) - 1
            i = blocks_num
            while i >= 0:  # (i = blocks_num i > = 0 i--):
                number = math.floor(int(blocks[i]))
                text = self._written_block(number)
                if text:
                    if number == 1 and i != 0:
                        text = self.complications[i][4]
                    elif number == 2 and i != 0:
                        text = self.complications[i][self._format]
                    elif number > 2 and number < 11 and i != 0:
                        text += " " + self.complications[i][3]
                    elif i != 0:
                        text += " " + self.complications[i][4]
                    items.append(text)
                i -= 1
            string = " و ".join(items)
        else:
            string = "صفر"
        return string

    def _written_block(self, number):
        """
        Spell sub block number of three digits max in Arabic idiom
        @param number: number Sub block number of three digits max you want to spell in Arabic idiom
        @type number: integer
        @return: The Arabic idiom that spells inserted sub block
        @rtype: String
        """
        items = []
        string = ""
        number = int(number)
        if number > 99:
            hundred = math.floor(number / 100) * 100
            number = number % 100

            if hundred == 200:
                items.append(self._individual[hundred][self._format])
            else:
                items.append(self._individual[hundred])
        if number == 2 or number == 12:
            items.append(self._individual[number][self._feminine][self._format])
        elif number < 20:
            items.append(self._individual[int(number)][self._feminine])
        else:
            ones = number % 10
            tens = math.floor(number / 10) * 10
            tens = int(tens)

            if ones == 2:
                items.append(self._individual[ones][self._feminine][self._format])
            elif ones > 0:
                items.append(self._individual[ones][self._feminine])
            items.append(self._individual[tens][self._format])

        if "" in items:
            items.remove("")
        string = " و ".join(items)
        return string


def text2number(text):
    """
    Convert arabic text into number, for example convert تسعة وعشرون = >29.

    Example:
        >>> text2number(u"خمسمئة وثلاث وعشرون")
        523

    @param text: input text
    @type text: unicode
    @return: number extracted from text
    @rtype: integer
    """
    # the result total is 0
    total = 0
    # the partial total for the three number
    partial = 0
    text = araby.strip_tashkeel(text)
    words = text.split(" ")
    # print words
    for word in words:
        if word and word != "واحد" and word[0] in ("و", "ف", "ل", "ب", "ك"):
            word = word[1:]
        if word != "واحد" and word.startswith("و"):
            word = word[1:]

        if word in nbconst.NUMBER_WORDS:
            actualnumber = nbconst.NUMBER_WORDS[word]
            if actualnumber % 1000 == 0:
                # the case of 1000 or 1 million
                if partial == 0:
                    partial = 1
                total += partial * actualnumber
                # re-initiate the partial total
                partial = 0
            else:
                partial += nbconst.NUMBER_WORDS[word]
    # add the final partial to total
    total += partial
    return total


def number2text(anumber):
    """
    Convert number to arabic words, for example convert 25 --> خمسة و عشرون

    Example:
        >>> number2text(523)
        خمسمئة وثلاث وعشرون

    @param anumber: input number
    @type anumber: int
    @return: number words
    @rtype: unicode
    """
    # test if the given type is numeric(float or int
    # if ok, convert it to string
    if type(anumber) is int or type(anumber) is float:
        anumber = str(anumber)
    # if the given type is str/unicode, test if it's a valid number
    elif type(anumber) is str or type(anumber) is unicode:
        try:
            a = float(anumber)
        except ValueError:
            return "صفر"
    # if the given number not a valid return 0
    else:
        return "صفر"
    arbn = ArNumbers()
    return arbn.int2str(anumber)

    return total


def vocalize_number(wordlist, syn_tags=""):
    """Vocalize a number words clause


    @param wordlist: words to vocalize
    @type wordlist: unicode list
    @param syn_tags: tags about the clause
    @type syn_tags: unicode
    @return: the vocalized wordlist.
    @rtype: unicode
    """
    newlist = []
    prefix = ""
    nextword = ""
    # we can pass tags to this number word
    tags = syn_tags
    if len(wordlist) == 1:
        word = wordlist[0]
        word_nm = araby.strip_tashkeel(word)
        key = word_nm
        voc = word
        # the first word can have prefixes
        if (
            word_nm
            and not wordlist
            and word_nm != "واحد"
            and word[0] in ("و", "ف", "ل", "ب", "ك")
        ):
            if word_nm[0] in ("ل", "ب", "ك"):
                tags += "مجرور"
            key = word[1:]
        elif word_nm != "واحد" and word_nm.startswith("و"):
            key = word_nm[1:]
        # تحذب بعض الكلمات لأنها تلتبس مع أسماء الأجزاء مثل خُمس وخمس
        if key in nbconst.NUMBER_WORDS and key not in (
            "عشر",
            "خمس",
            "سبع",
            "تسع",
            "خمسا",
            "سبعا",
            "تسعا",
            "عشرا",
            "ألفين",
            "عشرة",
            "صفر",
            "ألف",
        ):
            voc = prefix + nbconst.VOCALIZED_NUMBER_WORDS[key]["i"]
        return [
            voc,
        ]
    for i, word in enumerate(wordlist):
        # save the original word with possible harakat if exist
        # ~ word = wordlist[i]
        word_nm = araby.strip_tashkeel(word)
        key = word_nm
        # the first word can have prefixes
        if (
            i == 0
            and word_nm
            and word_nm != "واحد"
            and word[0] in ("و", "ف", "ل", "ب", "ك")
        ):
            if word_nm[0] in ("ل", "ب", "ك"):
                tags += "مجرور"
            key = word[1:]
        elif word_nm != "واحد" and word_nm.startswith("و"):
            key = word_nm[1:]
        if key in nbconst.NUMBER_WORDS:
            if word_nm.endswith("ين"):
                tags += "مجهول"  # إما مجرور أو منصوب
            elif word_nm.endswith("ان") or word_nm.endswith("ون"):
                tags += "مرفوع"
    pre_key = ""
    for i, word in enumerate(wordlist):
        # ~ word = wordlist[i]
        if i + 1 < len(wordlist):
            nextword = wordlist[i + 1]
        else:
            nextword = ""
        key = word
        # the first word can have prefixes
        if word and word != "واحد" and word[0] in ("و", "ف", "ل", "ب", "ك"):
            key = word[1:]
            prefix = word[0]
            if prefix in ("و", "ف", "ك"):
                prefix += "َ"
            elif prefix in ("ل", "ب"):
                prefix += "ِ"
        else:
            prefix = ""
        if key in nbconst.VOCALIZED_NUMBER_WORDS:
            voc = ""
            if nbconst.VOCALIZED_NUMBER_WORDS[key]["s"] == "*":
                voc = prefix + nbconst.VOCALIZED_NUMBER_WORDS[key]["i"]

            # مبني على النصب في حالة المركب العددي
            elif nextword == "عشر" or nextword == "عشرة":
                voc = prefix + nbconst.VOCALIZED_NUMBER_WORDS[key]["n"]
            # مبني على النصب في حالة المركب العددي
            elif key == "عشر" and pre_key in nbconst.NUMBER_TEN_MASCULIN_UNITS:
                voc = "عَشَرَ"
            elif key == "عشرة" and pre_key in nbconst.NUMBER_TEN_FEMININ_UNITS:
                voc = "عَشْرَةَ"
            elif "مرفوع" in tags:
                if nextword.startswith("و"):
                    voc = prefix + nbconst.VOCALIZED_NUMBER_WORDS[key]["r2"]
                else:
                    voc = prefix + nbconst.VOCALIZED_NUMBER_WORDS[key]["r"]
            elif "مجهول" in tags:
                voc = prefix + nbconst.VOCALIZED_NUMBER_WORDS[key]["i"]

            elif "مجرور" in tags:
                if nextword.startswith("و"):
                    voc = prefix + nbconst.VOCALIZED_NUMBER_WORDS[key]["j2"]
                else:
                    voc = prefix + nbconst.VOCALIZED_NUMBER_WORDS[key]["j"]
            # منصوب
            elif "منصوب" in tags:
                if nextword.startswith("و"):
                    voc = prefix + nbconst.VOCALIZED_NUMBER_WORDS[key]["n2"]
                else:
                    voc = prefix + nbconst.VOCALIZED_NUMBER_WORDS[key]["n"]
            else:
                voc = prefix + nbconst.VOCALIZED_NUMBER_WORDS[key]["i"]
            newlist.append(voc)
        else:
            newlist.append(prefix + key)
        pre_key = key
    return newlist


def is_unit(word):
    """
    return if the given word is a unit
    @param word: given word to be tested
    @type word: unicode
    @return: if word is a unit return True else False.
    @rtype: Boolean
    """
    return word in nbconst.UNIT_WORDS


def vocalize_unit(numeric, unit):
    """Vocalize a number words
    @param numeric: given number
    @type numeric: integer
    @param unit: unit to vocalize
    @type unit: unicode
    @return: the vocalized unit, or unit word if itsnt a unit word.
    @rtype: unicode
    """
    # detect tags
    # The given word is not a unit
    unit_nm = araby.strip_tashkeel(unit)
    if not is_unit(unit_nm):
        return unit
    tags = ""
    vocalizedunit = unit

    # العدد بين واحد واثنان يتطلب صفة للوحدة ويكون بعدها
    # هذه الحالة لا تبرمج

    if numeric >= 0 and numeric <= 2:
        return unit
    # الإضافة إلى تمييز مضاف  إليه مجرور مفرد
    # تممييز الألف والمئة والمليون والمليار
    # يتطلب إضافة إلى مفرد
    # مثلا ألف رجل
    elif numeric % 100 == 0 or numeric % 1000 == 0:
        tags = "SingleMajrour"
        vocalizedunit = nbconst.UNIT_WORDS[unit_nm]["a"]
    # العدد المفرد يتطلب
    # إضافة إلى الجمع
    elif numeric % 100 <= 10:
        tags += "Plural"
        vocalizedunit = nbconst.UNIT_WORDS[unit_nm]["p"]

    elif numeric % 100 < 100:
        tags += "SingleMansoub"
        vocalizedunit = nbconst.UNIT_WORDS[unit_nm]["n"]
    else:
        tags = ""
        vocalizedunit = nbconst.UNIT_WORDS[unit_nm]["i"]
    if not vocalizedunit:
        return "Error" + tags
    else:
        return vocalizedunit


def get_previous_tag(word):
    """Get the word tags
    @param word: given word
    @type word: unicode
    @return:word tag
    @rtype: unicode
    """
    word = araby.strip_tashkeel(word)
    # ~ tags = u''
    if word in nmconst.NOUN_NASEB_LIST:
        return "منصوب"
    elif word in nmconst.JAR_LIST:
        return "مجرور"
    elif word in nmconst.RAFE3_LIST:
        return "مرفوع"
    else:
        return ""


def extract_number_phrases(text):
    """
    Extract number words in a text.

    Example:
        >>> extract_number_phrases(u"وجدت خمسمئة وثلاثة وعشرين دينارا فاشتريت ثلاثة عشر دفترا")
        خمسمئة وثلاثة وعشرين
        ثلاثة عشر

    @param text: input text
    @type text: unicode
    @return: number words extracted from text
    @rtype: integer
    """
    phrases = []

    wordlist = araby.tokenize(text)  # text.split(' ')
    positions = detect_number_phrases_position(wordlist)

    for pos in positions:
        if len(pos) >= 2:
            if pos[0] <= len(wordlist) and pos[1] <= len(wordlist):
                phrases.append(" ".join(wordlist[pos[0] : pos[1] + 1]))
    return phrases


def extract_number_context(
    text,
):
    """
    Extract number words in a text within context.

    Example:
        >>> extract_number_context(u"وجدت خمسمئة وثلاثة وعشرين دينارا فاشتريت ثلاثة عشر دفترا")
        ‎وجدت، خمسمئة وثلاثة وعشرين، دينارا
        ‎فاشتريت، ثلاثة عشر ، دفتر

    @param text: input text
    @type text: unicode
    @return: number words extracted from text
    @rtype: integer
    """
    phrases = []
    wordlist = araby.tokenize(text)
    positions = detect_number_phrases_position(wordlist)

    for pos in positions:
        if len(pos) >= 2:
            if pos[0] <= len(wordlist) and pos[1] <= len(wordlist):
                if pos[0] - 1 >= 0:
                    prev = wordlist[pos[0] - 1]
                else:
                    prev = ""
                if pos[1] + 1 < len(wordlist):
                    nextword = wordlist[pos[1] + 1]
                else:
                    nextword = ""
                phrases.append(
                    (prev, " ".join(wordlist[pos[0] : pos[1] + 1]), nextword)
                )
    return phrases


def detect_number_phrases_position(wordlist):
    """
    Detect number words in a text and return positions of each phrase.
    @param wordlist: wordlist
    @type wordlist: unicode list
    @return: list of numbers clause positions [(start,end),(start2,end2),]
    @rtype: list of tuple
    """
    # ~ wordlist# = text.split(u' ')
    # print words
    phrases = []
    startnumber = -1
    endnumber = False
    # ~ taglist = []
    for i, word in enumerate(wordlist):
        # ~ word = wordlist[i]
        if i + 1 < len(wordlist):
            nextword = araby.strip_tashkeel(wordlist[i + 1])
        else:
            nextword = None
        # save the original word with possible harakat if exist
        word_nm = araby.strip_tashkeel(word)
        key = word_nm
        # the first word can have prefixes
        if (
            word_nm
            and not startnumber
            and word_nm != "واحد"
            and word_nm[0] in ("و", "ف", "ل", "ب", "ك")
        ):
            key = word_nm[1:]
        elif word_nm != "واحد" and word_nm.startswith("و"):
            key = word_nm[1:]
        if key in nbconst.NUMBER_WORDS or key.isnumeric():
            if key not in (
                "أحد",
                "إحدى",
                "اثنا",
                "اثني",
                "اثنتي",
                "اثنتا",
            ) or nextword in ("عشر", "عشرة"):
                if startnumber < 0:
                    startnumber = i
                endnumber = i
            # phrase.append(word)
        else:
            if startnumber >= 0:  # There are a previous number phrase.
                phrases.append((startnumber, endnumber))
            startnumber = -1
    # add the final phrases
    if startnumber >= 0:  # There are a previous number phrase.
        phrases.append((startnumber, endnumber))

    return phrases


def detect_numbers(wordlist):
    """
    Detect number words in a text and return a taglist as BIO.
    @param wordlist: wordlist
    @type wordlist: unicode list
    @return: list of tags BIO
    @rtype: list of unicode
    """
    # ~ phrases = []
    starts = False
    taglist = []

    for i, word in enumerate(wordlist):
        # ~ word = wordlist[i]
        if i + 1 < len(wordlist):
            nextword = araby.strip_tashkeel(wordlist[i + 1])
        else:
            nextword = None
        # save the original word with possible harakat if exist
        word_nm = araby.strip_tashkeel(word)
        key = word_nm
        # the first word can have prefixes
        if (
            word_nm
            and not starts
            and word_nm != "واحد"
            and word_nm[0] in ("و", "ف", "ل", "ب", "ك")
        ):
            key = word_nm[1:]
        elif word_nm != "واحد" and word_nm.startswith("و"):
            key = word_nm[1:]
        if key in nbconst.NUMBER_WORDS or key.isnumeric():
            if key not in (
                "أحد",
                "إحدى",
                "اثنا",
                "اثني",
                "اثنتي",
                "اثنتا",
            ) or nextword in ("عشر", "عشرة"):
                if not starts:
                    taglist.append("DB")
                    starts = True
                else:
                    taglist.append("DI")
            else:
                starts = False
                taglist.append("O")
        else:
            starts = False
            taglist.append("O")
    return taglist


def detect_number_words(text):
    """
    Detect number words in a text.

    Example:
        >>> detect_number_words(u"وجدت خمسمئة وثلاثة وعشرين دينارا")
        خمسمئة وثلاثة وعشرين

    @param text: input text
    @type text: unicode
    @return: number words extracted from text
    @rtype: integer
    """

    phrases_context = extract_number_context(text)
    for ph_con in phrases_context:
        if len(ph_con) >= 3:
            previous = ph_con[0]
            phrase = ph_con[1]
            nextword = ph_con[2]
            numberedwords = phrase
            numeric = text2number(numberedwords)
            tags = get_previous_tag(previous)
            wordlist = araby.strip_tashkeel(numberedwords).split(" ")
            vocalized = vocalize_number(wordlist, tags)
            # calcul  vocalization similarity:
            sim = araby.vocalized_similarity(numberedwords, vocalized)
            voc_unit = vocalize_unit(numeric, nextword)
            sim_unit = araby.vocalized_similarity(voc_unit, nextword)

            if sim < 0:
                # ~ print u'\t'.join([str(sim), u' '.join(numberedwords), vocalized,
                # ~ str(numeric), u' '.join([previous, phrase, nextword]),
                # ~ nextword, voc_unit, str(sim_unit)]).encode('utf8')
                print(
                    "\t".join([str(sim), " ".join(numberedwords), " ".join(vocalized)])
                )
                print(str(numeric), " ".join([previous, phrase, nextword]))
                print("\t".join([nextword, voc_unit, str(sim_unit)]))


def pre_tashkeel_number(wordlist):
    """
    Vocalized a number clauses in a text.
    @param wordlist: input text
    @type wordlist: unicode
    @return: wordlist with vocalized number clause
    @rtype: list
    """
    taglist = detect_numbers(wordlist)
    previous = ""
    vocalized_list = []
    chunk = []
    previous_tag = ""
    for word, tag in zip(wordlist, taglist):
        if tag in ("DB", "DI"):
            chunk.append(word)
        else:
            if chunk:
                # get the tag of previous word
                previous_tag = get_previous_tag(previous)
                vocalized = vocalize_number(chunk, previous_tag)
                vocalized_list.extend(vocalized)
                chunk = []
            vocalized_list.append(word)
            previous = word
    if chunk:
        vocalized = vocalize_number(chunk, previous_tag)
        vocalized_list.extend(vocalized)
        chunk = []
    return vocalized_list


def number2ordinal(anumber, feminin=False):
    """
    Convert number to arabic words in ordinal form, for example convert 25 --> الخامس والعشرون

      Example:
        >>> number2text(523)
        الخمسمئة والثالث والعشرون
    @param anumber: input number
    @type anumber: int
    @return: number words
    @rtype: unicode
    """
    # test if the given type is numeric(float or int
    # if ok, convert it to string
    a = 0
    if type(anumber) is int:
        anumber = str(anumber)
    # if the given type is str/unicode, test if it's a valid number
    elif type(anumber) is str or type(anumber) is unicode:
        try:
            a = int(anumber)
        except ValueError:
            return "صفر"
    # if the given number not a valid return 0
    else:
        return "الصفر"
    if a == 1:
        if feminin:
            return "الأولى"
        else:
            return "الأول"
    arbn = ArNumbers()
    arbn_str = arbn.int2str(anumber)
    # substitute numeral words to ordinal words
    # normalize waw
    arbn_str = arbn_str.replace("و ", "و")
    tokens = arbn_str.split(" ")
    # the first token:
    # if it's  unit, change it to ordinal
    # and add definate article
    new_list = []
    if feminin:
        ordinal_words = nbconst.UNITS_ORDINAL_WORDS_FEMININ
    else:
        ordinal_words = nbconst.UNITS_ORDINAL_WORDS
    if tokens:
        tok = tokens[0]
        if tok in ordinal_words:
            # ثلاث = ثالث
            tok = ordinal_words.get(tok, tok)
        # add definite article
        tok = "ال" + tok
        new_list.append(tok)

        for tok in tokens[1:]:
            # first strip first Waw
            if tok.startswith(araby.WAW):
                # strip waw
                tok = tok[1:]
                if tok in ordinal_words:
                    # ثلاث = ثالث
                    tok = ordinal_words.get(tok, tok)
                tok = "وال" + tok
            new_list.append(tok)
    # ajust الواحد at the end
    if new_list[-1:] == "الحادي":
        new_list.pop()
        new_list.append("الواحد")
    elif new_list[-1:] == "الحادية":
        new_list.pop()
        new_list.append("الواحدة")
    ordinal_string = " ".join(new_list)
    return ordinal_string


if __name__ == "__main__":
    # import number as ArabicNumberToLetters
    TEXTS = [
        "مليونان وألفان وإثنا عشر",
        "جاء مليونان وألفان وإثنا عشر",
        "وجدت خمسمئة وثلاث وعشرون دينارا",
        "خمسمئة وثلاث وعشرون دينارا",
        "وجدت خمسمئة وثلاثة وعشرين دينارا فاشتريت ثلاثة عشر دفترا",
        "لم أجد شيئا",
        "وجدت خمسمئة وثلاثة وعشرين دينارا فاشتريت ثلاثة عشر دفترا",
        "من ثلاثمئة وخمسين بلدا ",
        "من ثلاثمئة وخمسين بلدا ",
        "من أربعمئة وخمسين بلدا ",
        "السلام عليكم 2014",
    ]
    for txt in TEXTS:
        word_list = araby.tokenize(txt)
        positions_phrases = detect_number_phrases_position(word_list)
        nb_phrases = extract_number_phrases(txt)
        tag_list = detect_numbers(word_list)
        tashkeel = " ".join(pre_tashkeel_number(word_list))



#!/usr/bin/python
# -*- coding=utf-8 -*-
"""
Constants for named module
"""

RAFE3_LIST = set(
    [
        "أنه",
        "أنك",
        "أنها",
        "بأنها",
        "بأنه",
        "وأنها",
        "فأنها",
        "فأنه",
        "كأنه",
        "كأنها",
        # yahia alhadj
        "كان",
        "يكون",
        "كانت",
        "صار",
        "صارت",
        "يصير",
        "أمسى",
        "ليس",
        "ليست",
        "ظلّ",
        "ظلّت",
        "أضحى",
        "أضحت",
        "يضحي",
        "أصبح",
        "أصبحت",
        "يصبح",
        "بات",
        "باتت",
        "يبيت",
        "مازال",
        "لازال",
        "لايزال",
        "لازالت",
        "مايزال",
        "مازالت",
        "ماتزال",
        "مابرح",
        "مايبرح",
        "مابرحت",
        "ماانفك",
        "ماانفكّت",
        "ماينفك",
        "لاينفك",
        "مادام",
        "مادامت",
        "نعم",
        "بئس",
        "حبذا",
        # إضافي
        "هل",
        # u'من',
        "ما",
        "متى",
        "أين",
        "ماذا",
        "كيف",
        "أيان",
        # اسماء الإشارة بعد فاء الإستئناف
        "فهذا",
        "فذلك",
        "فتلك",
        "فهؤلاء",
        "فأولئك",
        "فذلكم",
        "فهذه",
        # ضمائر الرفع المنفصلة
        "هو",
        "هما",
        "هم",
        "هي",
        "هما",
        "هن",
        "أنت",
        "أنتما",
        "أنتم",
        "أنت",
        "أنتما",
        "أنتن",
        "أنا",
        "نحن",
        "إذ",
        # ------------
        # خاص بكتب التراث
        "قال",
        "أخبرنا",
        "أخبرني",
        "ثنا",
    ]
)

JAR_LIST = set(
    [
        "من",
        "عن",
        "إلى",
        "على",
        "في",
        "رب",
        "منذ",
        "مذ",
        "عدا",
        "خلا",
        "حاشا",
        "عند",
        "أمام",
        "وراء",
        "خلف",
        "مع",
        "قبل",
        "بعد",
        "تحت",
        "أي",
        "كلّ",
        "بعض",
        "غير",
        "سوى",
        "ليل",
        "شمال",
        "جنوب",
        "يمين",
        "شرق",
        "غرب",
        "شطر",
        "أسفل",
        "أعلى",
        "جنب",
        "جانب",
        "تلقاء",
        "قدام",
        "أعلى",
        "شهر",
        "سنة",
        "غروب",
        "شروق",
        "دون",
        "شهور",
        "يوم",
        "حين",
        "ساعة",
        "زمان",
        "أزمان",
        "أيام",
        "أوقات",
        "وقت",
        "لحظة",
        "خلال",
        "بدون",
        "أثناء",
        "ذات",
        "ذو",
        "ذوو",
        "ذوات",
        "ذوي",
        "بن",
        "ابن",
        "بنت",
        "بين",
        # صيغ واضحة الإضافة
        "أبو",
        "أخو",
        "بواسطة",
        "فَوْقَ",
        "مِنْ",
        "إِلَى",
        "رُبَّ",
        "عَلَى",
        "عَنْ",
        "فِي",
        "مِنْ",
        "عَمَّا",
        "حَتَّى",
        "مُنْذُ",
        "مُذْ",
        "فَإِلَى",
        "فَرُبَّ",
        "فَعَلَى",
        "فَعَنْ",
        "فَفِي",
        "فَمِنْ",
        "فَعَمَّا",
        "فَحَتَّى",
        "فَمُنْذُ",
        "فَمُذْ",
        "وَإِلَى",
        "وَرُبَّ",
        "وَعَلَى",
        "وَعَنْ",
        "وَفِي",
        "وَمِنْ",
        "وَعَمَّا",
        "وَحَتَّى",
        "وَمُنْذُ",
        "وَمُذْ",
    ]
)

NOUN_NASEB_LIST = set(
    [
        "أن",
        "إن",
        "فإن",
        "لأن",
        "كأن",
        "لكن",
        "ليت",
        "لعل",
        # vocalized factor
        "أَنَّ",
        "فَإَنَّ",
    ]
)

PROPER_NOUNS = [
    "عاصم",
    "جبريل",
    "أحمد",
]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import wave
from typing import Dict
from ovos_utils.log import LOG
from ovos_plugin_manager.templates.tts import TTS

from phoonnx.model_manager import TTSModelManager, TTSModelInfo
from phoonnx.voice import TTSVoice, SynthesisConfig


class PhoonnxTTSPlugin(TTS):
    """Interface to Phoonnx TTS."""
    engines = {}

    def __init__(self, config=None):
        """
        Initialize the PhoonnxTTSPlugin, prepare the model manager, and load an initial voice.
        
        Creates a TTSModelManager and attempts to refresh available voices; if refresh fails the manager falls back to loading voices. If a non-default voice is configured on the instance, that voice is loaded and cached in self.voices. Otherwise the language-specific default voice is determined, loaded, and stored in self.voices.
        
        Parameters:
            config (dict|None): Configuration passed to the base TTS initializer; may influence plugin behavior (e.g., selected voice or language).
        """
        super().__init__(config=config)
        self.model_manager = TTSModelManager()
        try:
            self.model_manager.refresh_voices()
        except Exception as exc:
            LOG.warning(f"Voice refresh failed; retrying with cached voices: {exc}")
            self.model_manager.load()

        self.voices: Dict[str, TTSVoice] = {}
        if self.voice and self.voice != "default":
            self.voices[self.voice] = self.get_model(self.voice)
        else:
            default = self.get_default_voice(self.lang)
            self.voices[default.voice_id] = self.get_model(default.voice_id)

    def get_default_voice(self, lang: str) -> TTSModelInfo:
        """
        Selects the default TTS model for the given language.
        
        Parameters:
        	lang (str): Language tag used to look up available voices (e.g., "en-US", "pt-PT").
        
        Returns:
        	TTSModelInfo: The first/default voice model info for the specified language.
        
        Raises:
        	ValueError: If no voices are available for the given language.
        """
        voices = self.model_manager.get_lang_voices(lang)
        if not voices:
            raise ValueError(f"No voices available for language: {lang}")
        return voices[0]

    def get_model(self, voice_id: str) -> TTSVoice:
        if voice_id in self.voices:
            return self.voices[voice_id]
        if voice_id not in self.model_manager.voices:
            raise Exception(f"Unknown voice: {voice_id}")
        LOG.debug(f"Using voice: {voice_id}")
        self.voices[voice_id] = self.model_manager.voices[voice_id].load()
        return self.voices[voice_id]

    def get_tts(self, sentence, wav_file, lang=None, voice=None):
        """
        Synthesize speech for a sentence and write the audio to the specified WAV file.
        
        Parameters:
            sentence (str): Text to synthesize.
            wav_file (str): Path to the output WAV file that will be written.
            lang (str, optional): Language override used to select the default voice when no `voice` is provided.
            voice (str, optional): Voice identifier override to select a specific model.
        
        Returns:
            tuple: (wav_file, phonemes) where `wav_file` is the path to the written WAV file and `phonemes` is `None` when phoneme output is not produced.
        """
        if voice:
            voice_info = self.model_manager.voices[voice]
            model = self.get_model(voice)
        else:
            voice_info = self.get_default_voice(lang or self.lang)
            model = self.get_model(voice_info.voice_id)

        synth_params = SynthesisConfig(
            enable_phonetic_spellings=self.config.get("enable_phonetic_spelling", True),
            add_diacritics=self.config.get("add_diacritics", voice_info.config.add_diacritics), # arabic and hebrew only
            noise_scale=self.config.get("noise-scale", voice_info.config.noise_scale),  # generator noise
            length_scale=self.config.get("length-scale", voice_info.config.length_scale),  # Phoneme length
            noise_w_scale=self.config.get("noise-w", voice_info.config.noise_w_scale)  # Phoneme width noise
        )
        with wave.open(wav_file, "wb") as wav_out:
            model.synthesize_wav(sentence, wav_out, synth_params)

        return wav_file, None


if __name__ == "__main__":
    utterance = "Guimarães é uma das mais importantes cidades históricas do país, estando o seu centro histórico inscrito na lista de Património Mundial da UNESCO desde 2001, o que a torna definitivamente num dos maiores centros turísticos da região. As suas ruas e monumentos respiram história e encantam quem a visita."
    #utterance = "Um arco-íris, também popularmente denominado arco-da-velha, é um fenômeno óptico e meteorológico que separa a luz do sol em seu espectro contínuo quando o sol brilha sobre gotículas de água suspensas no ar."

    tts = PhoonnxTTSPlugin()
    tts.get_tts(utterance, "tmiro-pt-PT.wav",
                voice="OpenVoiceOS/phoonnx_pt-PT_miro_tugaphone")
    tts.get_tts(utterance, "tdii-pt-PT.wav",
                voice="OpenVoiceOS/phoonnx_pt-PT_dii_tugaphone")
    tts.get_tts(utterance, "miro-pt-PT.wav",
                voice="OpenVoiceOS/pipertts_pt-PT_miro")
    tts.get_tts(utterance, "dii-pt-PT.wav",
                voice="OpenVoiceOS/pipertts_pt-PT_dii")


#!/usr/bin/python
# -*- coding=utf-8 -*-
"""
Stack module
@author: Taha Zerrouki
@contact: taha dot zerrouki at gmail dot com
@copyright: Arabtechies,  Arabeyes,   Taha Zerrouki
@license: GPL
@date:2010/03/01
@version: 0.1
"""


class Stack(object):
    """
    Stack class
    """

    def __init__(self, text=""):
        """
        create a stack
        """
        self.items = list(text)

    def push(self, item):
        """
        puch an item into the stack
        @param item: pushed item
        @type item: mixed
        @return: None
        @rtype: None
        """
        self.items.append(item)

    def pop(self):
        """
        pop an item from the stack
        @return: poped item
        @rtype: mixed
        """
        if not self.is_empty():
            return self.items.pop()
        else:
            return None

    def is_empty(self):
        """
        test if the stack is empty
        @return: True or False
        @rtype: boolean
        """
        return self.items == []



import click
from phoonnx.model_manager import TTSModelManager, TTSModelInfo
import requests


# --- Helper Function for CLI Output ---
def print_voice_info(voice: TTSModelInfo):
    """Prints detailed info about a single voice."""
    click.echo(f"  ID:          {voice.voice_id}")
    click.echo(f"  Language:    {voice.lang}")
    click.echo(f"  Engine:      {voice.engine.value}")
    click.echo(f"  Phoneme Type: {voice.phoneme_type}")
    click.echo(f"  Model URL:   {voice.model_url}")
    click.echo(f"  Config URL:  {voice.config_url}")
    click.echo("-" * 40)


# --- CLI Group ---
@click.group()
def cli():
    """CLI to manage and download phoonnx TTS voice models."""
    pass


# --- CLI Commands ---

@cli.command(name="update-cache")
@click.option("--no-clear", is_flag=True, help="Do not clear the existing cache before updating. Only adds new voices.")
def update_cache(no_clear):
    """
    Clears the local voice cache, fetches the latest voice lists from upstream
    sources (Piper, Mimic3, OpenVoiceOS), and saves them to the cache.
    """
    manager = TTSModelManager()

    if not no_clear:
        click.echo("Clearing local voice cache...")
        manager.clear()
    else:
        click.echo("Loading existing voice cache...")
        manager.load()

    click.echo("Fetching voice lists from upstream sources (this requires an internet connection)...")

    # Run the fetch methods to populate the cache
    try:
        manager.get_ovos_voice_list()
        click.echo("-> Fetched OpenVoiceOS voices.")
        manager.get_proxectonos_voice_list()
        click.echo("-> Fetched Proxectonos voices.")
        manager.get_phonikud_voice_list()
        click.echo("-> Fetched Phonikud voices.")
        manager.get_piper_voice_list()
        click.echo("-> Fetched Piper voices.")
        manager.get_mimic3_voice_list()
        click.echo("-> Fetched Mimic3 voices.")
    except requests.exceptions.RequestException as e:
        click.echo(f"\nError: Could not fetch voice lists due to a network or connection error.", err=True)
        click.echo(f"Details: {e}", err=True)
        return
    except Exception as e:
        click.echo(f"An unexpected error occurred while fetching voice lists: {e}", err=True)
        return

    manager.save()
    click.echo(f"\nCache updated successfully!")
    click.echo(f"Total voices available: {len(manager.all_voices)}")
    click.echo(f"Total languages supported: {len(manager.supported_langs)}")


@cli.command(name="list-langs")
def list_langs():
    """Lists all supported language codes available in the local cache."""
    manager = TTSModelManager()
    manager.load()

    if not manager.supported_langs:
        click.echo("No languages found. Run 'update-cache' first.")
        return

    click.echo(f"\nSupported Languages ({len(manager.supported_langs)}):")
    for l in manager.supported_langs:
        click.echo(f" - {l}")


@cli.command(name="list-voices")
@click.option("--lang", default=None, help="Filter voices by language code (e.g., 'en-US' or 'pt-PT').")
@click.option("--verbose", "-v", is_flag=True, help="Show detailed information for each voice.")
def list_voices(lang, verbose):
    """Lists all available voice models."""
    manager = TTSModelManager()
    manager.load()

    if not manager.all_voices:
        click.echo("No voices found in cache. Run 'update-cache' first.")
        return

    if lang:
        voices = manager.get_lang_voices(lang)
        click.echo(f"\nFound {len(voices)} voices for language '{lang}':")
    else:
        voices = manager.all_voices
        click.echo(f"\nFound {len(voices)} total voices:")

    if not voices:
        click.echo(f"No voices found for '{lang}'.")
        return

    for voice in voices:
        if verbose:
            print_voice_info(voice)
        else:
            click.echo(f"* {voice.voice_id} ({voice.lang})")


@cli.command(name="download")
@click.argument("voice_id", type=str)
def download_voice(voice_id):
    """
    Downloads the model, config, and token files for a specific VOICE_ID.
    The VOICE_ID must exist in the local cache (run update-cache first).
    """
    manager = TTSModelManager()
    manager.load()

    if voice_id not in manager.voices:
        click.echo(f"Error: Voice ID '{voice_id}' not found in cache.", err=True)
        click.echo("Hint: Run 'phoonnx_cli.py update-cache' first to fetch the list.", err=True)
        return

    voice_info = manager.voices[voice_id]

    try:
        click.echo(f"Attempting to download files for: {voice_id} ({voice_info.lang})")

        # Download config (model.json)
        click.echo("-> Downloading config...")
        voice_info.download_config()

        # Download tokens/phoneme map
        if voice_info.tokens_url:
            click.echo("-> Downloading tokens/phoneme map...")
            voice_info.download_phoneme_map()

        # Download model (model.onnx)
        click.echo("-> Downloading ONNX model (this may take a while)...")
        voice_info.download_model()

        click.echo(f"\nDownload complete. Files saved to: {voice_info.voice_path}")

    except requests.exceptions.RequestException as e:
        click.echo(f"\nDownload failed due to network error: {e}", err=True)
    except Exception as e:
        click.echo(f"\nAn unexpected error occurred during download: {e}", err=True)


if __name__ == "__main__":
    cli()


"""Test suite for phoonnx Arabic phonemizer (MantoqPhonemizer)."""
import unittest
from unittest.mock import patch
import sys
import os

from phoonnx.phonemizers.ar import MantoqPhonemizer
from phoonnx.phonemizers.base import BasePhonemizer
from phoonnx.config import Alphabet


class TestMantoqPhonemizer(unittest.TestCase):
    """Comprehensive unit tests for MantoqPhonemizer class.
    
    Testing Framework: unittest (Python standard library)
    Following existing project patterns from tests/test_util.py
    """

    def setUp(self):
        """Set up test fixtures before each test method."""
        self.phonemizer_mantoq = MantoqPhonemizer()
        self.phonemizer_ipa = MantoqPhonemizer(alphabet=Alphabet.IPA)

    def tearDown(self):
        """Clean up after each test method."""
        pass

    # Constructor and Initialization Tests
    def test_init_default_alphabet(self):
        """Test initialization with default BUCKWALTER alphabet."""
        phonemizer = MantoqPhonemizer()
        self.assertEqual(phonemizer.alphabet, Alphabet.BUCKWALTER)
        self.assertIsInstance(phonemizer, BasePhonemizer)

    def test_init_ipa_alphabet(self):
        """Test initialization with IPA alphabet."""
        phonemizer = MantoqPhonemizer(alphabet=Alphabet.IPA)
        self.assertEqual(phonemizer.alphabet, Alphabet.IPA)
        self.assertIsInstance(phonemizer, BasePhonemizer)

    def test_init_unicode_alphabet(self):
        """Test initialization with UNICODE alphabet."""
        with self.assertRaises(ValueError, msg="unsupported alphabet"):
            phonemizer = MantoqPhonemizer(alphabet=Alphabet.UNICODE)

    def test_init_inheritance_chain(self):
        """Test proper inheritance from BasePhonemizer."""
        self.assertIsInstance(self.phonemizer_mantoq, BasePhonemizer)
        self.assertTrue(hasattr(self.phonemizer_mantoq, 'alphabet'))
        self.assertTrue(hasattr(self.phonemizer_mantoq, 'phonemize_string'))
        # Verify super().__init__ was called properly
        self.assertIsNotNone(self.phonemizer_mantoq.alphabet)

    # Language Validation Tests (get_lang method)
    def test_get_lang_valid_arabic_codes(self):
        """Test get_lang with various valid Arabic language codes."""
        valid_codes = [
            "ar", "AR", "Ar", "aR",  # Case variations
            "ar-SA", "ar-EG", "ar-LB", "ar-JO", "ar-AE",  # Regional variants
            "ar_SA", "ar_EG", "ar_LB", "ar_JO", "ar_AE",  # Underscore variants
            "ara"  # "ISO 639-2 code"
        ]

        for code in valid_codes:
            with self.subTest(code=code):
                result = MantoqPhonemizer.get_lang(code)
                self.assertEqual(result, "ar", f"Failed to validate Arabic code: {code}")

    def test_get_lang_invalid_languages(self):
        """Test get_lang rejection of non-Arabic language codes."""
        invalid_codes = [
            "en", "fr", "es", "de", "it", "ru", "zh", "ja", "ko", "hi",
            "en-US", "fr-FR", "es-ES", "de-DE", "zh-CN", "ja-JP"
        ]

        for code in invalid_codes:
            with self.subTest(code=code):
                with self.assertRaises(ValueError, msg=f"Should reject non-Arabic code: {code}"):
                    MantoqPhonemizer.get_lang(code)

    def test_get_lang_edge_cases(self):
        """Test get_lang with edge case inputs."""
        edge_cases = [
            ("", "empty string"),
            (" ", "whitespace only"),
            ("ar ", "trailing space"),
            (" ar", "leading space"),
            ("  ar  ", "multiple spaces"),
            ("AR-", "incomplete region code"),
            ("_ar", "leading underscore"),
            ("ar_", "trailing underscore"),
            ("arabic", "full language name")  # TODO support this
        ]

        for case, description in edge_cases:
            with self.subTest(case=repr(case), desc=description):
                # Most edge cases should raise ValueError
                with self.assertRaises(ValueError, msg=f"Should reject edge case: {description}"):
                    MantoqPhonemizer.get_lang(case)

    def test_get_lang_invalid_types(self):
        """Test get_lang with invalid input types."""
        invalid_inputs = [
            (None, "None value"),
            (123, "integer"),
            (12.34, "float"),
            ([], "empty list"),
            (['ar'], "list with string"),
            ({}, "empty dict"),
            ({'lang': 'ar'}, "dict with lang"),
            (object(), "generic object"),
            (True, "boolean True"),
            (False, "boolean False")
        ]

        for invalid_input, description in invalid_inputs:
            with self.subTest(input=invalid_input, desc=description):
                with self.assertRaises((ValueError, TypeError, AttributeError)):
                    MantoqPhonemizer.get_lang(invalid_input)

    def test_get_lang_as_class_method(self):
        """Test get_lang called as class method."""
        result = MantoqPhonemizer.get_lang("ar")
        self.assertEqual(result, "ar")

    def test_get_lang_as_instance_method(self):
        """Test get_lang called as instance method."""
        result = self.phonemizer_mantoq.get_lang("ar")
        self.assertEqual(result, "ar")

    def test_get_lang_consistency_class_vs_instance(self):
        """Test that get_lang returns consistent results when called as class vs instance method."""
        test_codes = ["ar", "AR", "ar-SA", "ar_EG"]

        for code in test_codes:
            with self.subTest(code=code):
                class_result = MantoqPhonemizer.get_lang(code)
                instance_result = self.phonemizer_mantoq.get_lang(code)
                self.assertEqual(class_result, instance_result)

    # Basic Phonemization Tests with BUCKWALTER alphabet
    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_mantoq_basic(self, mock_mantoq):
        """Test basic phonemize_string functionality with BUCKWALTER alphabet."""
        mock_mantoq.return_value = ("مرحبا", ['m', 'a', 'r', 'h', 'a', 'b', 'a'])

        result = self.phonemizer_mantoq.phonemize_string("مرحبا", "ar")

        mock_mantoq.assert_called_once_with("مرحبا")
        self.assertEqual(result, "marhaba")
        self.assertIsInstance(result, str)

    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_mantoq_single_character(self, mock_mantoq):
        """Test phonemization of single Arabic character."""
        mock_mantoq.return_value = ("ب", ['b'])

        result = self.phonemizer_mantoq.phonemize_string("ب", "ar")

        mock_mantoq.assert_called_once_with("ب")
        self.assertEqual(result, "b")

    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_mantoq_with_word_separators(self, mock_mantoq):
        """Test phonemize_string with word separator tokens (_+_)."""
        mock_mantoq.return_value = ("مرحبا بالعالم", ['m', 'a', 'r', '_+_', 'h', 'a', 'b', 'a'])

        result = self.phonemizer_mantoq.phonemize_string("مرحبا بالعالم", "ar")

        mock_mantoq.assert_called_once_with("مرحبا بالعالم")
        self.assertEqual(result, "mar haba")
        self.assertIn(' ', result)
        self.assertEqual(result.count(' '), 1)

    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_mantoq_multiple_separators(self, mock_mantoq):
        """Test phonemize_string with multiple word separators."""
        mock_mantoq.return_value = ("text", ['a', '_+_', 'b', '_+_', 'c', '_+_', 'd'])

        result = self.phonemizer_mantoq.phonemize_string("كلام كثير", "ar")

        self.assertEqual(result, "a b c d")
        self.assertEqual(result.count(' '), 3)

    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_mantoq_consecutive_separators(self, mock_mantoq):
        """Test handling of consecutive word separators."""
        mock_mantoq.return_value = ("text", ['word', '_+_', '_+_', 'another'])

        result = self.phonemizer_mantoq.phonemize_string("test", "ar")

        # Should result in double space
        self.assertEqual(result, "word  another")

    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_mantoq_no_separators(self, mock_mantoq):
        """Test phonemize_string with no word separators."""
        mock_mantoq.return_value = ("كتاب", ['k', 'i', 't', 'a', 'b'])

        result = self.phonemizer_mantoq.phonemize_string("كتاب", "ar")

        self.assertEqual(result, "kitab")
        self.assertNotIn(' ', result)

    # Phonemization Tests with IPA alphabet
    @patch('phoonnx.phonemizers.ar.bw2ipa')
    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_ipa_basic(self, mock_mantoq, mock_bw2ipa):
        """Test phonemize_string with IPA alphabet."""
        mock_mantoq.return_value = ("مرحبا", ['m', 'a', 'r', 'h', 'a', 'b', 'a'])
        mock_bw2ipa.return_value = "marħaba"

        result = self.phonemizer_ipa.phonemize_string("مرحبا", "ar")

        mock_mantoq.assert_called_once_with("مرحبا")
        mock_bw2ipa.assert_called_once_with("marhaba")
        self.assertEqual(result, "marħaba")

    @patch('phoonnx.phonemizers.ar.bw2ipa')
    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_ipa_with_separators(self, mock_mantoq, mock_bw2ipa):
        """Test IPA phonemization preserving word separators."""
        mock_mantoq.return_value = ("مرحبا بالعالم", ['m', 'a', 'r', '_+_', 'h', 'a', 'b', 'a'])
        mock_bw2ipa.return_value = "mar ħaba"

        result = self.phonemizer_ipa.phonemize_string("مرحبا بالعالم", "ar")

        mock_mantoq.assert_called_once_with("مرحبا بالعالم")
        mock_bw2ipa.assert_called_once_with("mar haba")
        self.assertEqual(result, "mar ħaba")

    @patch('phoonnx.phonemizers.ar.bw2ipa')
    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_ipa_complex_phonemes(self, mock_mantoq, mock_bw2ipa):
        """Test IPA phonemization with complex Arabic phonemes."""
        mock_mantoq.return_value = ("الشمس", ['?', 'a', 'l', 'S', 'a', 'm', 's'])
        mock_bw2ipa.return_value = "ʔaʃːams"  # IPA representation with gemination

        result = self.phonemizer_ipa.phonemize_string("الشمس", "ar")

        mock_bw2ipa.assert_called_once_with("?alSams")
        self.assertEqual(result, "ʔaʃːams")

    # Edge Cases and Error Handling
    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_empty_text(self, mock_mantoq):
        """Test phonemize_string with empty text."""
        mock_mantoq.return_value = ("", [])

        result = self.phonemizer_mantoq.phonemize_string("", "ar")

        mock_mantoq.assert_called_once_with("")
        self.assertEqual(result, "")

    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_whitespace_only(self, mock_mantoq):
        """Test phonemize_string with whitespace-only text."""
        mock_mantoq.return_value = ("   ", [' ', ' ', ' '])

        result = self.phonemizer_mantoq.phonemize_string("   ", "ar")

        self.assertEqual(result, "   ")

    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_newlines_and_tabs(self, mock_mantoq):
        """Test phonemize_string with newlines and tabs."""
        mock_mantoq.return_value = ("text\nwith\ttabs",
                                    ['t', 'e', 'x', 't', '\n', 'w', 'i', 't', 'h', '\t', 't', 'a', 'b', 's'])

        result = self.phonemizer_mantoq.phonemize_string("مرحبا\nبالعالم\tاليوم", "ar")

        self.assertIn('\n', result)
        self.assertIn('\t', result)

    def test_phonemize_string_default_language_parameter(self):
        """Test phonemize_string with default language parameter (should be 'ar')."""
        with patch('phoonnx.phonemizers.ar.mantoq') as mock_mantoq:
            mock_mantoq.return_value = ("text", ['t', 'e', 's', 't'])

            result = self.phonemizer_mantoq.phonemize_string("مرحبا")

            mock_mantoq.assert_called_once_with("مرحبا")
            self.assertEqual(result, "test")

    def test_phonemize_string_invalid_language(self):
        """Test phonemize_string with invalid language code."""
        invalid_langs = ["en", "fr", "es", "de", "zh", "ja", "invalid"]

        for lang in invalid_langs:
            with self.subTest(lang=lang):
                with self.assertRaises(ValueError):
                    self.phonemizer_mantoq.phonemize_string("test", lang)

    # Exception Handling and Error Propagation
    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_mantoq_exception(self, mock_mantoq):
        """Test phonemize_string when mantoq raises an exception."""
        mock_mantoq.side_effect = Exception("Mantoq processing error")

        with self.assertRaises(Exception) as context:
            self.phonemizer_mantoq.phonemize_string("مرحبا", "ar")

        self.assertIn("Mantoq processing error", str(context.exception))

    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_mantoq_import_error(self, mock_mantoq):
        """Test phonemize_string when mantoq has import issues."""
        mock_mantoq.side_effect = ImportError("mantoq library not found")

        with self.assertRaises(ImportError):
            self.phonemizer_mantoq.phonemize_string("مرحبا", "ar")

    @patch('phoonnx.phonemizers.ar.bw2ipa')
    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_bw2ipa_exception(self, mock_mantoq, mock_bw2ipa):
        """Test phonemize_string when bw2ipa raises an exception."""
        mock_mantoq.return_value = ("text", ['t', 'e', 's', 't'])
        mock_bw2ipa.side_effect = Exception("bw2ipa translation error")

        with self.assertRaises(Exception) as context:
            self.phonemizer_ipa.phonemize_string("مرحبا", "ar")

        self.assertIn("bw2ipa translation error", str(context.exception))

    @patch('phoonnx.phonemizers.ar.bw2ipa')
    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_bw2ipa_import_error(self, mock_mantoq, mock_bw2ipa):
        """Test phonemize_string when bw2ipa has import issues."""
        mock_mantoq.return_value = ("text", ['t', 'e', 's', 't'])
        mock_bw2ipa.side_effect = ImportError("bw2ipa library not found")

        with self.assertRaises(ImportError):
            self.phonemizer_ipa.phonemize_string("مرحبا", "ar")

    # Data Format Validation Tests
    @patch('phoonnx.phonemizers.ar.bw2ipa')
    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_bw2ipa_return_types(self, mock_mantoq, mock_bw2ipa):
        """Test phonemize_string when bw2ipa returns various types."""
        mock_mantoq.return_value = ("text", ['t', 'e', 's', 't'])

        test_cases = [
            (None, None),
            ("", ""),
            ("valid_ipa", "valid_ipa"),
            (123, 123)  # Should handle non-string returns
        ]

        for return_value, expected in test_cases:
            with self.subTest(return_value=return_value):
                mock_bw2ipa.return_value = return_value
                result = self.phonemizer_ipa.phonemize_string("مرحبا", "ar")
                self.assertEqual(result, expected)

    # Comprehensive Arabic Text Tests
    def test_phonemize_string_arabic_test_cases_from_main(self):
        """Test phonemize_string with the Arabic samples from the main block."""
        arabic_test_cases = [
            ("مرحبا بالعالم", "Hello world greeting"),
            ("ذهب الطالب إلى المكتبة لقراءة كتاب عن تاريخ الأندلس.", "Complex sentence with various elements"),
            ("الشمس", "Sun - tests sun letter assimilation"),
            ("فيل", "Elephant - tests long vowels"),
            ("يوم", "Day - tests glides"),
            ("سور", "Wall - tests long vowels"),
            ("لو", "If - tests glides")
        ]

        with patch('phoonnx.phonemizers.ar.mantoq') as mock_mantoq, \
                patch('phoonnx.phonemizers.ar.bw2ipa') as mock_bw2ipa:
            for arabic_text, description in arabic_test_cases:
                with self.subTest(text=arabic_text, desc=description):
                    # Mock realistic phoneme output
                    mock_phonemes = list(arabic_text.replace(' ', '_+_'))
                    mock_mantoq.return_value = (arabic_text, mock_phonemes)
                    mock_bw2ipa.return_value = f"ipa_{arabic_text}"

                    # Test BUCKWALTER alphabet
                    result_mantoq = self.phonemizer_mantoq.phonemize_string(arabic_text, "ar")
                    self.assertIsInstance(result_mantoq, str)

                    # Test IPA alphabet
                    result_ipa = self.phonemizer_ipa.phonemize_string(arabic_text, "ar")
                    self.assertIsInstance(result_ipa, str)
                    self.assertTrue(result_ipa.startswith("ipa_"))

    def test_phonemize_string_additional_arabic_samples(self):
        """Test phonemize_string with additional Arabic text samples."""
        additional_samples = [
            ("كتاب", "Book - simple word"),
            ("مدرسة", "School"),
            ("بيت", "House"),
            ("ماء", "Water"),
            ("نار", "Fire"),
            ("هواء", "Air"),
            ("أرض", "Earth"),
            ("قمر", "Moon"),
            ("نجم", "Star"),
            ("بحر", "Sea")
        ]

        with patch('phoonnx.phonemizers.ar.mantoq') as mock_mantoq:
            for arabic_text, description in additional_samples:
                with self.subTest(text=arabic_text, desc=description):
                    mock_phonemes = list(arabic_text)
                    mock_mantoq.return_value = (arabic_text, mock_phonemes)

                    result = self.phonemizer_mantoq.phonemize_string(arabic_text, "ar")
                    self.assertIsInstance(result, str)
                    mock_mantoq.assert_called_with(arabic_text)

    # Special Characters and Mixed Content Tests
    def test_phonemize_string_mixed_content(self):
        """Test phonemize_string with mixed Arabic and other content."""
        mixed_cases = [
            ("مرحبا123", "Arabic with numbers"),
            ("مرحبا!", "Arabic with exclamation"),
            ("مرحبا؟", "Arabic with Arabic question mark"),
            ("مرحبا?", "Arabic with Latin question mark"),
            ("مرحبا، كيف حالك؟", "Arabic with punctuation"),
            ("مرحبا (test)", "Arabic with parentheses"),
            ("مرحبا@gmail.com", "Arabic with email-like text"),
            ("123 مرحبا 456", "Arabic surrounded by numbers"),
            ("Hello مرحبا World", "Arabic mixed with English"),
            ("مرحبا - Welcome", "Arabic with dash and English")
        ]

        with patch('phoonnx.phonemizers.ar.mantoq') as mock_mantoq:
            for text, description in mixed_cases:
                with self.subTest(text=repr(text), desc=description):
                    mock_mantoq.return_value = ("normalized", ['t', 'e', 's', 't'])

                    result = self.phonemizer_mantoq.phonemize_string(text, "ar")
                    self.assertIsInstance(result, str)
                    mock_mantoq.assert_called_with(text)

    # Performance and Stress Tests
    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_very_long_text(self, mock_mantoq):
        """Test phonemize_string with very long Arabic text."""
        long_text = "مرحبا بالعالم الجميل " * 100  # ~2000 characters
        long_phonemes = ['m', 'a', 'r'] * 500  # Simulate long phoneme output

        mock_mantoq.return_value = ("normalized", long_phonemes)

        result = self.phonemizer_mantoq.phonemize_string(long_text, "ar")
        self.assertIsInstance(result, str)
        self.assertEqual(len(result), 1500)  # 3 chars * 500 repetitions

    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_empty_phonemes_list(self, mock_mantoq):
        """Test phonemize_string when mantoq returns empty phonemes list."""
        mock_mantoq.return_value = ("text", [])

        result = self.phonemizer_mantoq.phonemize_string("مرحبا", "ar")
        self.assertEqual(result, "")

    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_single_separator(self, mock_mantoq):
        """Test phonemize_string with only separator tokens."""
        mock_mantoq.return_value = ("text", ['_+_'])

        result = self.phonemizer_mantoq.phonemize_string("   ", "ar")
        self.assertEqual(result, " ")

    # Type Validation Tests
    def test_phonemize_string_parameter_type_validation(self):
        """Test phonemize_string with various invalid parameter types."""
        invalid_text_types = [
            (None, "None value"),
            (123, "integer"),
            (12.34, "float"),
            ([], "empty list"),
            ({}, "empty dict"),
            (object(), "generic object"),
            (True, "boolean True"),
            (False, "boolean False")
        ]

        for invalid_text, description in invalid_text_types:
            with self.subTest(text_type=type(invalid_text).__name__, desc=description):
                with self.assertRaises((TypeError, AttributeError, ValueError)):
                    self.phonemizer_mantoq.phonemize_string(invalid_text, "ar")

    # Word Boundary and Separator Preservation Tests
    @patch('phoonnx.phonemizers.ar.mantoq')
    def test_phonemize_string_word_boundary_edge_cases(self, mock_mantoq):
        """Test edge cases in word boundary handling."""
        test_cases = [
            (["a", "_+_", "b"], "a b", 1, "Basic separator"),
            (["a", "_+_", "b", "_+_", "c"], "a b c", 2, "Multiple separators"),
            (["_+_", "starts", "with", "_+_", "sep"], " startswith sep", 2, "Leading separator"),
            (["ends", "with", "_+_"], "endswith ", 1, "Trailing separator"),
            (["_+_", "_+_", "multiple", "_+_", "_+_"], "  multiple  ", 4, "Multiple consecutive"),
            (["no", "separators", "here"], "noseparatorshere", 0, "No separators"),
            (["_+_"], " ", 1, "Only separator"),
            ([], "", 0, "Empty phonemes")
        ]

        for phonemes, expected, expected_spaces, description in test_cases:
            with self.subTest(phonemes=phonemes, desc=description):
                mock_mantoq.return_value = ("text", phonemes)

                result = self.phonemizer_mantoq.phonemize_string("test", "ar")
                self.assertEqual(result, expected, f"Unexpected result for: {description}")
                self.assertEqual(result.count(' '), expected_spaces, f"Wrong space count for: {description}")

    # Alphabet-specific Behavior Tests
    def test_alphabet_specific_processing_paths(self):
        """Test that different alphabet settings trigger different processing paths."""
        with patch('phoonnx.phonemizers.ar.mantoq') as mock_mantoq, \
                patch('phoonnx.phonemizers.ar.bw2ipa') as mock_bw2ipa:
            mock_mantoq.return_value = ("text", ['t', 'e', 's', 't'])

            # Test BUCKWALTER path (should not call bw2ipa)
            mock_bw2ipa.reset_mock()
            mantoq_result = self.phonemizer_mantoq.phonemize_string("مرحبا", "ar")
            mock_bw2ipa.assert_not_called()
            self.assertEqual(mantoq_result, "test")

            # Test IPA path (should call bw2ipa)
            mock_bw2ipa.return_value = "ipa_test"
            ipa_result = self.phonemizer_ipa.phonemize_string("مرحبا", "ar")
            mock_bw2ipa.assert_called_once_with("test")
            self.assertEqual(ipa_result, "ipa_test")

            # Results should be different
            self.assertNotEqual(mantoq_result, ipa_result)

    def test_alphabet_attribute_consistency(self):
        """Test that alphabet attribute remains consistent throughout object lifecycle."""
        # Test BUCKWALTER phonemizer
        self.assertEqual(self.phonemizer_mantoq.alphabet, Alphabet.BUCKWALTER)
        with patch('phoonnx.phonemizers.ar.mantoq') as mock_mantoq:
            mock_mantoq.return_value = ("text", [])
            self.phonemizer_mantoq.phonemize_string("test", "ar")
            # Alphabet should remain unchanged after phonemization
            self.assertEqual(self.phonemizer_mantoq.alphabet, Alphabet.BUCKWALTER)

        # Test IPA phonemizer
        self.assertEqual(self.phonemizer_ipa.alphabet, Alphabet.IPA)
        with patch('phoonnx.phonemizers.ar.mantoq') as mock_mantoq, \
                patch('phoonnx.phonemizers.ar.bw2ipa') as mock_bw2ipa:
            mock_mantoq.return_value = ("text", [])
            mock_bw2ipa.return_value = ""
            self.phonemizer_ipa.phonemize_string("test", "ar")
            # Alphabet should remain unchanged after phonemization
            self.assertEqual(self.phonemizer_ipa.alphabet, Alphabet.IPA)


class TestMantoqPhonemizerIntegration(unittest.TestCase):
    """Integration tests for MantoqPhonemizer.
    
    These tests check the integration between MantoqPhonemizer and its dependencies,
    though they still use mocks to avoid requiring actual external libraries.
    """

    def setUp(self):
        """Set up integration test fixtures."""
        self.phonemizer_mantoq = MantoqPhonemizer()
        self.phonemizer_ipa = MantoqPhonemizer(alphabet=Alphabet.IPA)

    def test_full_pipeline_mantoq_alphabet(self):
        """Test complete phonemization pipeline for BUCKWALTER alphabet."""
        with patch('phoonnx.phonemizers.ar.mantoq') as mock_mantoq:
            # Simulate realistic mantoq output
            mock_mantoq.return_value = (
                "مرحبا بالعالم",
                ['m', 'a', 'r', 'H', 'a', 'b', 'a', '_+_', 'b', 'i', 'l', 'E', 'a', 'l', 'a', 'm']
            )

            result = self.phonemizer_mantoq.phonemize_string("مرحبا بالعالم", "ar")

            # Verify full pipeline execution
            mock_mantoq.assert_called_once_with("مرحبا بالعالم")
            self.assertEqual(result, "marHaba bilEalam")
            self.assertIn(' ', result)

    def test_full_pipeline_ipa_alphabet(self):
        """Test complete phonemization pipeline for IPA alphabet."""
        with patch('phoonnx.phonemizers.ar.mantoq') as mock_mantoq, \
                patch('phoonnx.phonemizers.ar.bw2ipa') as mock_bw2ipa:
            # Simulate realistic mantoq output
            mock_mantoq.return_value = (
                "مرحبا بالعالم",
                ['m', 'a', 'r', 'H', 'a', 'b', 'a', '_+_', 'b', 'i', 'l', 'E', 'a', 'l', 'a', 'm']
            )
            # Simulate realistic bw2ipa output
            mock_bw2ipa.return_value = "marħaba bilʕaːlam"

            result = self.phonemizer_ipa.phonemize_string("مرحبا بالعالم", "ar")

            # Verify full pipeline execution
            mock_mantoq.assert_called_once_with("مرحبا بالعالم")
            mock_bw2ipa.assert_called_once_with("marHaba bilEalam")
            self.assertEqual(result, "marħaba bilʕaːlam")

    def test_error_propagation_through_pipeline(self):
        """Test that errors are properly propagated through the processing pipeline."""
        # Test mantoq error propagation
        with patch('phoonnx.phonemizers.ar.mantoq') as mock_mantoq:
            mock_mantoq.side_effect = RuntimeError("Mantoq processing failed")

            with self.assertRaises(RuntimeError) as context:
                self.phonemizer_mantoq.phonemize_string("مرحبا", "ar")

            self.assertIn("Mantoq processing failed", str(context.exception))

        # Test bw2ipa error propagation  
        with patch('phoonnx.phonemizers.ar.mantoq') as mock_mantoq, \
                patch('phoonnx.phonemizers.ar.bw2ipa') as mock_bw2ipa:
            mock_mantoq.return_value = ("text", ['t', 'e', 's', 't'])
            mock_bw2ipa.side_effect = RuntimeError("bw2ipa translation failed")

            with self.assertRaises(RuntimeError) as context:
                self.phonemizer_ipa.phonemize_string("مرحبا", "ar")

            self.assertIn("bw2ipa translation failed", str(context.exception))

    @unittest.skipIf(True, "Real integration test - enable when testing with actual libraries")
    def test_real_arabic_phonemization_samples(self):
        """Real integration test with actual Arabic text samples.
        
        This test should be enabled when testing with actual mantoq and bw2ipa libraries.
        Currently skipped to avoid dependency on external libraries in unit tests.
        """
        # Test cases from the main block of the original file
        test_cases = [
            "مرحبا بالعالم",
            "ذهب الطالب إلى المكتبة لقراءة كتاب عن تاريخ الأندلس.",
            "الشمس",
            "فيل",
            "يوم",
            "سور",
            "لو"
        ]

        for arabic_text in test_cases:
            with self.subTest(text=arabic_text):
                # Test with BUCKWALTER alphabet
                result_mantoq = self.phonemizer_mantoq.phonemize_string(arabic_text, "ar")
                self.assertIsInstance(result_mantoq, str)
                self.assertGreater(len(result_mantoq), 0)

                # Test with IPA alphabet
                result_ipa = self.phonemizer_ipa.phonemize_string(arabic_text, "ar")
                self.assertIsInstance(result_ipa, str)
                self.assertGreater(len(result_ipa), 0)

                # Results should be different (BUCKWALTER vs IPA)
                self.assertNotEqual(result_mantoq, result_ipa)


if __name__ == '__main__':
    # Create a test suite with comprehensive output
    unittest.main(verbosity=2, buffer=True)



from phoonnx.phonemizers.base import BasePhonemizer
from phoonnx.config import Alphabet


class TugaphonePhonemizer(BasePhonemizer):

    def __init__(self):
        from tugaphone import TugaPhonemizer
        self.tuga = TugaPhonemizer()
        super().__init__(Alphabet.IPA)

    @classmethod
    def get_lang(cls, target_lang: str) -> str:
        """
        Validates and returns the closest supported language code.

        Args:
            target_lang (str): The language code to validate.

        Returns:
            str: The validated language code.

        Raises:
            ValueError: If the language code is unsupported.
        """
        # this check is here only to throw an exception if invalid language is provided
        return cls.match_lang(target_lang, ["pt-PT", "pt-BR", "pt-AO", "pt-MZ", "pt-TL"])

    def phonemize_string(self, text: str, lang: str) -> str:
        lang = self.get_lang(lang)
        # TODO - support regional dialects
        return self.tuga.phonemize_sentence(text, lang)



if __name__ == "__main__":

    pho = TugaphonePhonemizer()

    sentences = [
        "O gato dorme.",
        "Tu falas português muito bem.",
        "O comboio chegou à estação.",
        "A menina comeu o pão todo.",
        "Vou pôr a manteiga no frigorífico.",
        "Ele está a trabalhar no escritório.",
        "Choveu muito ontem à noite.",
        "A rapariga comprou um telemóvel novo.",
        "Vamos tomar um pequeno-almoço.",
        "O carro ficou sem gasolina."
    ]

    for s in sentences:
        print(s)
        for code in ["pt-PT", "pt-BR", "pt-AO", "pt-MZ", "pt-TL"]:
            print(f"{code} → {pho.phonemize_string(s, code)}")
        print("######")



from phoonnx.phonemizers.base import BasePhonemizer
from phoonnx.config import Alphabet


class PersianPhonemizer(BasePhonemizer):
    """https://github.com/de-mh/persian_phonemizer"""
    def __init__(self, alphabet=Alphabet.IPA):
        from persian_phonemizer import Phonemizer
        assert alphabet in [Alphabet.ERAAB, Alphabet.IPA]
        output_format = "IPA" if alphabet == Alphabet.IPA else 'eraab'
        self.g2p = Phonemizer(output_format)
        super().__init__(alphabet)

    @classmethod
    def get_lang(cls, target_lang: str) -> str:
        """
        Validates and returns the closest supported language code.

        Args:
            target_lang (str): The language code to validate.

        Returns:
            str: The validated language code.

        Raises:
            ValueError: If the language code is unsupported.
        """
        # this check is here only to throw an exception if invalid language is provided
        return cls.match_lang(target_lang, ["fa"])

    def phonemize_string(self, text: str, lang: str = "fa") -> str:
        """
        """
        lang = self.get_lang(lang)
        return self.g2p.phonemize(text)


if __name__ == "__main__":
    text = "دوچرخه جدید علی گم شد."

    pho = PersianPhonemizer()
    lang = "fa"

    print(f"\n--- Getting phonemes for '{text}' ---")
    phonemes_cotovia = pho.phonemize(text, lang)
    print(f"  Phonemes: {phonemes_cotovia}")



# START_VERSION_BLOCK
VERSION_MAJOR = 0
VERSION_MINOR = 5
VERSION_BUILD = 2
VERSION_ALPHA = 1
# END_VERSION_BLOCK

VERSION_STR = f"{VERSION_MAJOR}.{VERSION_MINOR}.{VERSION_BUILD}"
if VERSION_ALPHA:
    VERSION_STR += f"a{VERSION_ALPHA}"



from typing import Optional, Tuple

import numpy as np

from .vad import SileroVoiceActivityDetector


def trim_silence(
    audio_array: np.ndarray,
    detector: SileroVoiceActivityDetector,
    threshold: float = 0.2,
    samples_per_chunk=480,
    sample_rate=16000,
    keep_chunks_before: int = 2,
    keep_chunks_after: int = 2,
) -> Tuple[float, Optional[float]]:
    """Returns the offset/duration of trimmed audio in seconds"""
    offset_sec: float = 0.0
    duration_sec: Optional[float] = None
    first_chunk: Optional[int] = None
    last_chunk: Optional[int] = None
    seconds_per_chunk: float = samples_per_chunk / sample_rate

    chunk = audio_array[:samples_per_chunk]
    audio_array = audio_array[samples_per_chunk:]
    chunk_idx: int = 0

    # Determine main block of speech
    while len(audio_array) > 0:
        prob = detector(chunk, sample_rate=sample_rate)
        is_speech = prob >= threshold

        if is_speech:
            if first_chunk is None:
                # First speech
                first_chunk = chunk_idx
            else:
                # Last speech so far
                last_chunk = chunk_idx

        chunk = audio_array[:samples_per_chunk]
        audio_array = audio_array[samples_per_chunk:]
        chunk_idx += 1

    if (first_chunk is not None) and (last_chunk is not None):
        first_chunk = max(0, first_chunk - keep_chunks_before)
        last_chunk = min(chunk_idx, last_chunk + keep_chunks_after)

        # Compute offset/duration
        offset_sec = first_chunk * seconds_per_chunk
        last_sec = (last_chunk + 1) * seconds_per_chunk
        duration_sec = last_sec - offset_sec

    return offset_sec, duration_sec



from phoonnx.config import Alphabet
from phoonnx.phonemizers.base import BasePhonemizer


class PhonikudPhonemizer(BasePhonemizer):

    def __init__(self):
        from phonikud import phonemize
        self.g2p = phonemize
        super().__init__(Alphabet.IPA)

    @classmethod
    def get_lang(cls, target_lang: str) -> str:
        """
        Validates and returns the closest supported language code.

        Args:
            target_lang (str): The language code to validate.

        Returns:
            str: The validated language code.

        Raises:
            ValueError: If the language code is unsupported.
        """
        # this check is here only to throw an exception if invalid language is provided
        return cls.match_lang(target_lang, ["he"])

    def phonemize_string(self, text: str, lang: str = "he") -> str:
        """
        """
        lang = self.get_lang(lang)
        return self.g2p(text)


if __name__ == "__main__":
    # text = "מתכת יקרה"
    text = 'שָׁלוֹם עוֹלָם'

    pho = PhonikudPhonemizer()
    lang = "he"

    print(f"\n--- Getting phonemes for '{text}' ---")
    # text = pho.add_diacritics(text, lang)
    phonemes = pho.phonemize(text, lang)
    print(f"  Phonemes: {phonemes}")
    # --- Getting phonemes for 'שָׁלוֹם עוֹלָם' ---
    #   Phonemes: [('ʃalˈom ʔolˈam', '.', True)]



import unittest
from unittest.mock import patch, MagicMock
from datetime import date
import sys
import os

from phoonnx.util import (
    normalize, _get_number_separators, _normalize_number_word,
    _normalize_dates_and_times, _normalize_word_hyphen_digit,
    _normalize_units, _normalize_word, is_fraction,
    pronounce_date, pronounce_time, CONTRACTIONS, TITLES, UNITS
)

class TestUtilFunctions(unittest.TestCase):
    """Comprehensive test suite for util.py functions."""

    def setUp(self):
        """Set up test fixtures."""
        self.mock_rbnf_engine = MagicMock()
        self.mock_rbnf_engine.format_number.return_value.text = "formatted number"

    def test_get_number_separators_default(self):
        """Test _get_number_separators with default languages."""
        decimal, thousands = _get_number_separators("en")
        self.assertEqual(decimal, '.')
        self.assertEqual(thousands, ',')

        decimal, thousands = _get_number_separators("en-US")
        self.assertEqual(decimal, '.')
        self.assertEqual(thousands, ',')

    def test_get_number_separators_european(self):
        """Test _get_number_separators with European languages."""
        for lang in ["pt", "es", "fr", "de"]:
            decimal, thousands = _get_number_separators(lang)
            self.assertEqual(decimal, ',')
            self.assertEqual(thousands, '.')

        # Test with full language codes
        decimal, thousands = _get_number_separators("pt-PT")
        self.assertEqual(decimal, ',')
        self.assertEqual(thousands, '.')

    def test_is_fraction_valid(self):
        """Test is_fraction with valid fractions."""
        self.assertTrue(is_fraction("1/2"))
        self.assertTrue(is_fraction("3/4"))
        self.assertTrue(is_fraction("10/20"))
        self.assertTrue(is_fraction("0/1"))

    def test_is_fraction_invalid(self):
        """Test is_fraction with invalid inputs."""
        self.assertFalse(is_fraction("1.5"))
        self.assertFalse(is_fraction("1/2/3"))
        self.assertFalse(is_fraction("a/b"))
        self.assertFalse(is_fraction("1/"))
        self.assertFalse(is_fraction("/2"))
        self.assertFalse(is_fraction("no_fraction"))
        self.assertFalse(is_fraction(""))

    def test_is_fraction_edge_cases(self):
        """Test is_fraction with edge cases."""
        self.assertFalse(is_fraction("1/2.5"))
        self.assertFalse(is_fraction("1.0/2"))
        self.assertFalse(is_fraction("1/-2"))
        self.assertFalse(is_fraction("-1/2"))

    @patch('phoonnx.util.pronounce_number')
    @patch('phoonnx.util.is_numeric')
    def test_normalize_number_word_simple_integer(self, mock_is_numeric, mock_pronounce):
        """Test _normalize_number_word with simple integers."""
        mock_is_numeric.return_value = True
        mock_pronounce.return_value = "twenty three"

        result = _normalize_number_word("23", "en", None)
        mock_pronounce.assert_called_with(23, lang="en")
        self.assertEqual(result, "twenty three")

    @patch('phoonnx.util.pronounce_number')
    @patch('phoonnx.util.is_numeric')
    def test_normalize_number_word_with_punctuation(self, mock_is_numeric, mock_pronounce):
        """Test _normalize_number_word preserves punctuation."""
        mock_is_numeric.return_value = True
        mock_pronounce.return_value = "twenty three"

        result = _normalize_number_word("23!", "en", None)
        mock_pronounce.assert_called_with(23, lang="en")
        self.assertEqual(result, "twenty three!")

    @patch('phoonnx.util.pronounce_fraction')
    def test_normalize_number_word_fraction(self, mock_pronounce_fraction):
        """Test _normalize_number_word with fractions."""
        mock_pronounce_fraction.return_value = "one half"

        with patch('phoonnx.util.is_fraction', return_value=True):
            result = _normalize_number_word("1/2", "en", None)
            mock_pronounce_fraction.assert_called_with("1/2", "en")
            self.assertEqual(result, "one half")

    @patch('phoonnx.util.pronounce_fraction')
    def test_normalize_number_word_fraction_with_punctuation(self, mock_pronounce_fraction):
        """Test _normalize_number_word with fractions and punctuation."""
        mock_pronounce_fraction.return_value = "three quarters"

        with patch('phoonnx.util.is_fraction', return_value=True):
            result = _normalize_number_word("3/4.", "en", None)
            mock_pronounce_fraction.assert_called_with("3/4", "en")
            self.assertEqual(result, "three quarters.")

    @patch('phoonnx.util.is_numeric')
    def test_normalize_number_word_european_decimal(self, mock_is_numeric):
        """Test _normalize_number_word with European decimal separator."""
        mock_is_numeric.side_effect = lambda x: x in ["1.2", "1,2"]

        with patch('phoonnx.util.pronounce_number') as mock_pronounce:
            mock_pronounce.return_value = "one point two"
            result = _normalize_number_word("1,2", "pt", None)
            mock_pronounce.assert_called_with(1.2, lang="pt")
            self.assertEqual(result, "one point two")

    @patch('phoonnx.util.is_numeric')
    def test_normalize_number_word_thousands_separator(self, mock_is_numeric):
        """Test _normalize_number_word with thousands separator."""
        mock_is_numeric.side_effect = lambda x: x in ["1234", "1,234"]

        with patch('phoonnx.util.pronounce_number') as mock_pronounce:
            mock_pronounce.return_value = "one thousand two hundred thirty four"
            result = _normalize_number_word("1,234", "en", None)
            mock_pronounce.assert_called_with(1234, lang="en")
            self.assertEqual(result, "one thousand two hundred thirty four")

    @patch('phoonnx.util.is_numeric')
    def test_normalize_number_word_complex_european_format(self, mock_is_numeric):
        """Test _normalize_number_word with complex European format (123.456,78)."""
        mock_is_numeric.side_effect = lambda x: x == "123456.78"

        with patch('phoonnx.util.pronounce_number') as mock_pronounce:
            mock_pronounce.return_value = "one hundred twenty three thousand four hundred fifty six point seven eight"
            _normalize_number_word("123.456,78", "pt", None)
            mock_pronounce.assert_called_with(123456.78, lang="pt")

    def test_normalize_number_word_rbnf_fallback(self):
        """Test _normalize_number_word RBNF fallback for digits."""
        mock_engine = MagicMock()
        mock_engine.format_number.return_value.text = "twenty three"

        with patch('phoonnx.util.is_numeric', return_value=False):
            result = _normalize_number_word("23", "en", mock_engine)
            mock_engine.format_number.assert_called_once()
            self.assertEqual(result, "twenty three")

    def test_normalize_number_word_no_change(self):
        """Test _normalize_number_word when no normalization is needed."""
        with patch('phoonnx.util.is_numeric', return_value=False):
            result = _normalize_number_word("hello", "en", None)
            self.assertEqual(result, "hello")

    @patch('phoonnx.util.nice_date')
    def test_pronounce_date(self, mock_nice_date):
        """Test pronounce_date function."""
        mock_nice_date.return_value = "January first, twenty twenty five"
        test_date = date(2025, 1, 1)

        result = pronounce_date(test_date, "en")
        mock_nice_date.assert_called_with(test_date, "en")
        self.assertEqual(result, "January first, twenty twenty five")

    @patch('phoonnx.util.nice_time')
    def test_pronounce_time_valid(self, mock_nice_time):
        """Test pronounce_time with valid military time."""
        mock_nice_time.return_value = "three fifteen PM"

        result = pronounce_time("15h15", "en")
        mock_nice_time.assert_called_once()
        self.assertEqual(result, "three fifteen PM")

    def test_pronounce_time_invalid(self):
        """Test pronounce_time with invalid time format."""
        result = pronounce_time("invalid", "en")
        self.assertEqual(result, "invalid")

    def test_pronounce_time_edge_case(self):
        """Test pronounce_time with edge cases."""
        result = pronounce_time("25h70", "en")
        # Should handle gracefully and return modified string
        self.assertIn(" ", result)

    def test_normalize_word_hyphen_digit(self):
        """Test _normalize_word_hyphen_digit function."""
        test_cases = [
            ("sub-23", "sub 23"),
            ("pre-10", "pre 10"),
            ("word-123", "word 123"),
            ("no-hyphen", "no-hyphen"),  # no digit after hyphen
            ("just-text", "just-text"),  # no digit
            #  ("123-456", "123-456"),     # no word before hyphen TODO Fix this one, should be pronounced number by number
        ]

        for input_text, expected in test_cases:
            with self.subTest(input_text=input_text):
                result = _normalize_word_hyphen_digit(input_text)
                self.assertEqual(result, expected)

    @patch('phoonnx.util.pronounce_number')
    def test_normalize_units_symbolic(self, mock_pronounce):
        """Test _normalize_units with symbolic units."""
        mock_pronounce.return_value = "twenty five"
        result = _normalize_units("25°C", "en")
        mock_pronounce.assert_called_with(25.0, "en")
        self.assertIn("twenty five", result)
        self.assertIn("degrees celsius", result)

    @patch('phoonnx.util.pronounce_number')
    def test_normalize_units_alphanumeric(self, mock_pronounce):
        """Test _normalize_units with alphanumeric units."""
        mock_pronounce.return_value = "five"

        result = _normalize_units("5kg", "en")
        mock_pronounce.assert_called_with(5.0, "en")
        self.assertIn("five", result)
        self.assertIn("kilograms", result)

    def test_normalize_units_unsupported_language(self):
        """Test _normalize_units with unsupported language."""
        result = _normalize_units("25°C", "unsupported")
        self.assertEqual(result, "25°C")  # Should remain unchanged

    @patch('phoonnx.util.pronounce_number')
    def test_normalize_units_european_format(self, mock_pronounce):
        """Test _normalize_units with European number format."""
        mock_pronounce.return_value = "vinte e cinco vírgula cinco"

        _normalize_units("25,5kg", "pt")
        mock_pronounce.assert_called_with(25.5, "pt")

    def test_normalize_word_contractions(self):
        """Test _normalize_word with contractions."""
        result = _normalize_word("can't", "en", None)
        self.assertEqual(result, "can not")

        result = _normalize_word("I'm", "en", None)
        self.assertEqual(result, "I am")

    def test_normalize_word_titles(self):
        """Test _normalize_word with titles."""
        result = _normalize_word("Dr.", "en", None)
        self.assertEqual(result, "Doctor")

        result = _normalize_word("Prof.", "en", None)
        self.assertEqual(result, "Professor")

    def test_normalize_word_multilingual_titles(self):
        """Test _normalize_word with titles in different languages."""
        result = _normalize_word("Sr.", "es", None)
        self.assertEqual(result, "Señor")

        result = _normalize_word("M.", "fr", None)
        self.assertEqual(result, "Monsieur")

    @patch('phoonnx.util._normalize_number_word')
    def test_normalize_word_delegates_numbers(self, mock_normalize_number):
        """Test _normalize_word delegates to _normalize_number_word."""
        mock_normalize_number.return_value = "twenty three"

        result = _normalize_word("23", "en", None)
        mock_normalize_number.assert_called_with("23", "en", None)
        self.assertEqual(result, "twenty three")

    def test_normalize_word_no_change(self):
        """Test _normalize_word when no normalization is needed."""
        result = _normalize_word("hello", "en", None)
        self.assertEqual(result, "hello")

    @patch('phoonnx.util.nice_time')
    def test_normalize_dates_and_times_military_time(self, mock_nice_time):
        """Test _normalize_dates_and_times with military time."""
        mock_nice_time.return_value = "three fifteen PM"

        result = _normalize_dates_and_times("Meeting at 15h15", "en")
        self.assertIn("three fifteen PM", result)

    def test_normalize_dates_and_times_am_pm_preprocessing(self):
        """Test _normalize_dates_and_times with AM/PM preprocessing."""
        result = _normalize_dates_and_times("Meeting at 3pm", "en")
        self.assertIn("P M", result)

        result = _normalize_dates_and_times("Call at 9am", "en")
        self.assertIn("A M", result)

    @patch('phoonnx.util.pronounce_date')
    def test_normalize_dates_and_times_date_parsing(self, mock_pronounce_date):
        """Test _normalize_dates_and_times with date parsing."""
        mock_pronounce_date.return_value = "March eighth, twenty twenty five"

        result = _normalize_dates_and_times("Due on 08/03/2025", "en-US", "MDY")
        mock_pronounce_date.assert_called_once()
        self.assertIn("March eighth, twenty twenty five", result)

    def test_normalize_dates_and_times_invalid_date(self):
        """Test _normalize_dates_and_times with invalid date."""
        # Should handle gracefully and not crash
        result = _normalize_dates_and_times("Due on 32/13/2025", "en")
        self.assertIn("32/13/2025", result)  # Should remain unchanged

    def test_normalize_dates_and_times_ambiguous_date_dmy(self):
        """Test _normalize_dates_and_times with ambiguous date using DMY format."""
        with patch('phoonnx.util.pronounce_date') as mock_pronounce_date:
            mock_pronounce_date.return_value = "May fifteenth, twenty twenty five"
            _normalize_dates_and_times("Due on 15/05/2025", "en", "DMY")
            mock_pronounce_date.assert_called_once()

    def test_normalize_dates_and_times_year_detection(self):
        """Test _normalize_dates_and_times year detection logic."""
        with patch('phoonnx.util.pronounce_date') as mock_pronounce_date:
            mock_pronounce_date.return_value = "formatted date"

            # Test 4-digit year at beginning
            _normalize_dates_and_times("2025/03/15", "en")

            # Test 4-digit year at end
            _normalize_dates_and_times("15/03/2025", "en")

    @patch('unicode_rbnf.RbnfEngine')
    @patch('phoonnx.util._normalize_word')
    def test_normalize_main_function(self, mock_normalize_word, mock_rbnf_engine):
        """Test main normalize function integration."""
        mock_normalize_word.side_effect = lambda w, lang, engine: w.upper()
        mock_rbnf_engine.for_language.return_value = self.mock_rbnf_engine

        result = normalize("hello world", "en")
        self.assertEqual(result, "HELLO WORLD")

    @patch('phoonnx.util._normalize_dates_and_times')
    @patch('phoonnx.util._normalize_word_hyphen_digit')
    @patch('phoonnx.util._normalize_units')
    def test_normalize_date_format_selection(self, mock_normalize_units,
                                             mock_normalize_word_hyphen_digit,
                                             mock_normalize_dates):
        """Test normalize function date format selection."""
        # Test for en-US, which should use MDY
        normalize("The date is 08/03/2025", "en-US")
        mock_normalize_dates.assert_called_with("The date is 08/03/2025", "en-US", "MDY")

        # Test for en-GB, which should use DMY
        normalize("The date is 08/03/2025", "en-GB")
        mock_normalize_dates.assert_called_with("The date is 08/03/2025", "en-GB", "DMY")

    @patch('unicode_rbnf.RbnfEngine')
    def test_normalize_rbnf_engine_error_handling(self, mock_rbnf_engine):
        """Test normalize function handles RBNF engine creation errors."""
        mock_rbnf_engine.for_language.side_effect = Exception("RBNF error")

        # Should not crash when RBNF engine fails to initialize
        result = normalize("test", "unsupported-lang")
        self.assertIsInstance(result, str)

    def test_normalize_empty_string(self):
        """Test normalize with empty string."""
        result = normalize("", "en")
        self.assertEqual(result, "")

    def test_normalize_whitespace_only(self):
        """Test normalize with whitespace only."""
        result = normalize("   ", "en")
        self.assertEqual(result, "")

    def test_normalize_single_word(self):
        """Test normalize with single word."""
        with patch('phoonnx.util._normalize_word') as mock_normalize_word:
            mock_normalize_word.return_value = "normalized"
            normalize("word", "en")
            mock_normalize_word.assert_called_with("word", "en", unittest.mock.ANY)

    def test_contractions_dictionary_completeness(self):
        """Test that CONTRACTIONS dictionary is properly structured."""
        self.assertIn("en", CONTRACTIONS)
        self.assertIsInstance(CONTRACTIONS["en"], dict)
        self.assertGreater(len(CONTRACTIONS["en"]), 1)  # Should have contractions

        # Test some specific contractions
        self.assertEqual(CONTRACTIONS["en"]["can't"], "can not")
        self.assertEqual(CONTRACTIONS["en"]["I'm"], "I am")

    def test_titles_dictionary_completeness(self):
        """Test that TITLES dictionary is properly structured."""
        for lang in ["en", "ca", "es", "pt", "gl", "fr", "it", "nl", "de"]:
            if lang in TITLES:
                self.assertIsInstance(TITLES[lang], dict)
                self.assertIn("Dr.", TITLES[lang])

    def test_units_dictionary_completeness(self):
        """Test that UNITS dictionary is properly structured."""
        for lang in ["en", "pt", "es", "fr", "de"]:
            if lang in UNITS:
                self.assertIsInstance(UNITS[lang], dict)
                if "%" in UNITS[lang]:
                    self.assertIn("%", UNITS[lang])
                if "°" in UNITS[lang]:
                    self.assertIn("°", UNITS[lang])

    def test_data_integrity_contractions(self):
        """Test data integrity of contractions."""
        for _lang, contractions in CONTRACTIONS.items():
            for contraction, expansion in contractions.items():
                self.assertIsInstance(contraction, str)
                self.assertIsInstance(expansion, str)
                self.assertGreater(len(contraction), 0)
                self.assertGreater(len(expansion), 0)

    def test_data_integrity_titles(self):
        """Test data integrity of titles."""
        for _lang, titles in TITLES.items():
            for title, expansion in titles.items():
                self.assertIsInstance(title, str)
                self.assertIsInstance(expansion, str)
                self.assertGreater(len(title), 0)
                self.assertGreater(len(expansion), 0)

    def test_data_integrity_units(self):
        """Test data integrity of units."""
        for _lang, units in UNITS.items():
            for unit, expansion in units.items():
                self.assertIsInstance(unit, str)
                self.assertIsInstance(expansion, str)
                self.assertGreater(len(unit), 0)
                self.assertGreater(len(expansion), 0)

    def test_error_handling_fraction_pronunciation(self):
        """Test error handling in fraction pronunciation."""
        with patch('ovos_number_parser.pronounce_fraction', side_effect=Exception("Test error")), \
                patch('phoonnx.util.is_fraction', return_value=True):
            result = _normalize_number_word("1/2", "en", None)
            self.assertEqual(result, "1/2")  # Should return original on error

    def test_error_handling_number_pronunciation(self):
        """Test error handling in number pronunciation."""
        with patch('phoonnx.util.pronounce_number', side_effect=Exception("Test error")), \
                patch('phoonnx.util.is_numeric', return_value=True):
            result = _normalize_number_word("123", "en", None)
            self.assertEqual(result, "123")  # Should return original on error

    def test_error_handling_rbnf_pronunciation(self):
        """Test error handling in RBNF pronunciation."""
        mock_engine = MagicMock()
        mock_engine.format_number.side_effect = Exception("RBNF error")

        with patch('phoonnx.util.is_numeric', return_value=False):
            result = _normalize_number_word("123", "en", mock_engine)
            self.assertEqual(result, "123")  # Should return original on error

    def test_complex_integration_scenario(self):
        """Test complex integration scenario with multiple normalizations."""
        text = "Dr. Smith said I can't attend the 3pm meeting on 15/03/2025, it's 25°C outside"

        with patch('phoonnx.util._normalize_dates_and_times') as mock_dates:
            mock_dates.return_value = text
            with patch('phoonnx.util._normalize_units') as mock_units:
                mock_units.return_value = text
                with patch('phoonnx.util._normalize_word') as mock_word:
                    mock_word.side_effect = lambda w, lang, engine: f"NORM_{w}"

                    result = normalize(text, "en")

                    # Verify all normalization steps were called
                    mock_dates.assert_called_once()
                    mock_units.assert_called_once()
                    self.assertIn("NORM_", result)

    def test_edge_case_multiple_separators(self):
        """Test edge cases with multiple separators in numbers."""
        test_cases = [
            ("1.234.567,89", "pt"),  # Multiple thousands separators
            ("1,234,567.89", "en"),  # Multiple thousands separators
            ("1.2.3", "en"),  # Ambiguous format
        ]

        for test_input, lang in test_cases:
            with self.subTest(input=test_input, lang=lang):
                # Should not crash
                result = _normalize_number_word(test_input, lang, None)
                self.assertIsInstance(result, str)

    def test_performance_large_text(self):
        """Test performance with large text input."""
        large_text = "Dr. Smith " * 1000  # Repeat to create large text

        # Should complete in reasonable time without crashing
        result = normalize(large_text, "en")
        self.assertIsInstance(result, str)

    def test_unicode_handling(self):
        """Test handling of unicode characters."""
        unicode_text = "café naïve résumé"

        result = normalize(unicode_text, "en")
        self.assertIsInstance(result, str)
        # Should preserve unicode characters when no normalization applies
        self.assertIn("café", result)

    def test_normalize_word_case_sensitivity(self):
        """Test _normalize_word case sensitivity."""
        # Contractions should be case-sensitive
        result = _normalize_word("CAN'T", "en", None)
        self.assertEqual(result, "CAN'T")  # Should remain unchanged

        result = _normalize_word("can't", "en", None)
        self.assertEqual(result, "can not")

    def test_normalize_dates_complex_patterns(self):
        """Test _normalize_dates_and_times with complex date patterns."""
        # Test leap year
        with patch('phoonnx.util.pronounce_date') as mock_pronounce:
            mock_pronounce.return_value = "February twenty ninth, twenty twenty four"
            result = _normalize_dates_and_times("Meeting on 29/02/2024", "en", "DMY")
            self.assertIn("February twenty ninth", result)

    def test_normalize_units_spacing_variations(self):
        """Test _normalize_units with various spacing patterns."""
        with patch('ovos_number_parser.pronounce_number') as mock_pronounce:
            mock_pronounce.return_value = "twenty five"

            # Test with space
            result = _normalize_units("25 kg", "en")
            self.assertIn("twenty five", result)

            # Test without space
            result = _normalize_units("25kg", "en")
            self.assertIn("twenty five", result)

    def test_normalize_multiple_time_formats(self):
        """Test _normalize_dates_and_times with multiple time formats."""
        text = "Meeting at 14h30 and call at 9am"

        with patch('phoonnx.util.nice_time') as mock_nice_time:
            mock_nice_time.side_effect = ["two thirty PM", "nine A M"]
            result = _normalize_dates_and_times(text, "en")
            self.assertIn("two thirty PM", result)
            self.assertIn("A M", result)

    def test_normalize_fraction_edge_cases(self):
        """Test is_fraction and fraction normalization with edge cases."""
        # Test fraction with zero
        self.assertTrue(is_fraction("0/1"))
        self.assertTrue(is_fraction("1/0"))  # Mathematical invalid but syntactically valid

        # Test large numbers
        self.assertTrue(is_fraction("999/1000"))

    def test_normalize_number_word_float_conversion(self):
        """Test _normalize_number_word float vs int conversion logic."""
        with patch('phoonnx.util.pronounce_number') as mock_pronounce, \
                patch('phoonnx.util.is_numeric', return_value=True):
            mock_pronounce.return_value = "five"

            # Integer case
            result = _normalize_number_word("5", "en", None)
            mock_pronounce.assert_called_with(5, lang="en")  # int(5)

            # Float case
            _normalize_number_word("5.0", "en", None)
            mock_pronounce.assert_called_with(5.0, lang="en")  # float(5.0)

    def test_normalize_multilingual_comprehensive(self):
        """Test normalize function with comprehensive multilingual examples."""
        test_cases = [
            ("Hola Dr. García", "es", "Hola Doctor García"),
            ("Bonjour M. Dupont", "fr", "Bonjour Monsieur Dupont"),
            ("Olá Sr. Silva", "pt", "Olá Senhor Silva"),
        ]

        for text, lang, _expected_partial in test_cases:
            with self.subTest(text=text, lang=lang):
                result = normalize(text, lang)
                # Just check that title expansion occurred
                self.assertNotEqual(result, text)

    def test_normalize_units_priority_handling(self):
        """Test _normalize_units handles overlapping unit symbols correctly."""
        # Test that longer units are matched first (m vs mL)
        with patch('ovos_number_parser.pronounce_number') as mock_pronounce:
            mock_pronounce.return_value = "five"

            result = _normalize_units("5mL", "en")
            self.assertIn("milliliters", result)
            self.assertNotIn("meters", result)


class TestDataStructureIntegrity(unittest.TestCase):
    """Test the integrity and completeness of data structures."""

    def test_contractions_comprehensive_coverage(self):
        """Test that contractions cover common English patterns."""
        if "en" in CONTRACTIONS:
            en_contractions = CONTRACTIONS["en"]

            # Test modal verbs
            if "won't" in en_contractions:
                self.assertIn("won't", en_contractions)
            if "can't" in en_contractions:
                self.assertIn("can't", en_contractions)
            if "shouldn't" in en_contractions:
                self.assertIn("shouldn't", en_contractions)

            # Test complex contractions
            if "wouldn't've" in en_contractions:
                self.assertIn("wouldn't've", en_contractions)
            if "you'd've" in en_contractions:
                self.assertIn("you'd've", en_contractions)

    def test_units_comprehensive_coverage(self):
        """Test that units cover major measurement categories."""
        if "en" in UNITS:
            en_units = UNITS["en"]

            # Temperature
            if "°C" in en_units:
                self.assertIn("°C", en_units)
            if "°F" in en_units:
                self.assertIn("°F", en_units)

            # Currency
            if "$" in en_units:
                self.assertIn("$", en_units)
            if "€" in en_units:
                self.assertIn("€", en_units)
            if "£" in en_units:
                self.assertIn("£", en_units)

            # Distance
            if "km" in en_units:
                self.assertIn("km", en_units)
            if "m" in en_units:
                self.assertIn("m", en_units)
            if "ft" in en_units:
                self.assertIn("ft", en_units)

    def test_titles_professional_coverage(self):
        """Test that titles cover professional and social titles."""
        if "en" in TITLES:
            en_titles = TITLES["en"]

            if "Dr." in en_titles:
                self.assertIn("Dr.", en_titles)
            if "Prof." in en_titles:
                self.assertIn("Prof.", en_titles)
            if "Mr." in en_titles:
                self.assertIn("Mr.", en_titles)

    def test_consistency_across_languages(self):
        """Test consistency of common elements across languages."""
        common_units = ["€", "%", "°"]

        for lang in ["en", "pt", "es", "fr", "de"]:
            if lang in UNITS:
                for unit in common_units:
                    if unit in UNITS[lang]:
                        self.assertIn(unit, UNITS[lang],
                                      f"Unit '{unit}' missing from {lang}")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
import csv
import dataclasses
import itertools
import json
import logging
import os
from collections import Counter
from dataclasses import dataclass
from multiprocessing import JoinableQueue, Process, Queue
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Tuple, Any, Set, Union, Callable

import click
from phoonnx.util import normalize
from phoonnx.config import PhonemeType, get_phonemizer, Alphabet
from phoonnx.phonemizers import Phonemizer
from phoonnx.phoneme_ids import (
    phonemes_to_ids, DEFAULT_IPA_PHONEME_ID_MAP, DEFAULT_PAD_TOKEN,
    DEFAULT_BOS_TOKEN, DEFAULT_EOS_TOKEN, DEFAULT_BLANK_WORD_TOKEN
)
from phoonnx_train.norm_audio import cache_norm_audio, make_silence_detector
from tqdm import tqdm
from phoonnx.version import VERSION_STR

_LOGGER = logging.getLogger("preprocess")

# Base phoneme map
DEFAULT_SPECIAL_PHONEME_ID_MAP: Dict[str, int] = {
    DEFAULT_PAD_TOKEN: 0,
    DEFAULT_BOS_TOKEN: 1,
    DEFAULT_EOS_TOKEN: 2,
    DEFAULT_BLANK_WORD_TOKEN: 3,
}
MAX_PHONEMES = 256
# -----------------------------------------------------------------------------

@dataclass
class Utterance:
    """Represents a single utterance in the dataset."""
    text: str
    audio_path: Path
    speaker: Optional[str] = None
    speaker_id: Optional[int] = None
    phonemes: Optional[List[str]] = None
    phoneme_ids: Optional[List[int]] = None
    audio_norm_path: Optional[Path] = None
    audio_spec_path: Optional[Path] = None

    def asdict(self) -> Dict[str, Any]:
        """Custom asdict to handle Path objects for JSON serialization."""
        data = dataclasses.asdict(self)
        for key, value in data.items():
            if isinstance(value, Path):
                data[key] = str(value)
        return data


class PathEncoder(json.JSONEncoder):
    """JSON encoder for Path objects."""

    def default(self, o: Any) -> Union[str, Any]:
        """
        Converts Path objects to strings for serialization.

        Args:
            o: The object to serialize.

        Returns:
            The serialized string representation or the default JSON serialization.
        """
        if isinstance(o, Path):
            return str(o)
        return super().default(o)


def get_text_casing(casing: str) -> Callable[[str], str]:
    """
    Returns a function to apply text casing based on a string name.

    Args:
        casing: The name of the casing function ('lower', 'upper', 'casefold', or 'ignore').

    Returns:
        A callable function (str) -> str.
    """
    if casing == "lower":
        return str.lower
    if casing == "upper":
        return str.upper
    if casing == "casefold":
        return str.casefold
    return lambda s: s


@dataclass
class PreprocessorConfig:
    """Dataclass to hold all runtime configuration, mimicking argparse.Namespace."""
    input_dir: Path
    output_dir: Path
    language: str
    sample_rate: int
    cache_dir: Path
    max_workers: int
    single_speaker: bool
    speaker_id: Optional[int]
    phoneme_type: PhonemeType
    alphabet: Alphabet
    phonemizer_model: str
    text_casing: str
    dataset_name: Optional[str]
    audio_quality: Optional[str]
    skip_audio: bool
    debug: bool
    add_diacritics: bool


def ljspeech_dataset(config: PreprocessorConfig) -> Iterable[Utterance]:
    """
    Generator for LJSpeech-style dataset.
    Loads metadata and resolves audio file paths.

    Args:
        config: The configuration object containing dataset parameters.

    Yields:
        Utterance: A fully populated Utterance object.
    """
    dataset_dir = config.input_dir
    metadata_path = dataset_dir / "metadata.csv"
    if not metadata_path.exists():
        _LOGGER.error(f"Missing metadata file: {metadata_path}")
        return

    wav_dirs: List[Path] = [dataset_dir / "wav", dataset_dir / "wavs"]

    with open(metadata_path, "r", encoding="utf-8") as csv_file:
        reader = csv.reader(csv_file, delimiter="|")
        for row in reader:
            if len(row) < 2:
                _LOGGER.warning(f"Skipping malformed row: {row}")
                continue

            filename: str = row[0]
            text: str = row[-1]
            speaker: Optional[str] = None

            if not config.single_speaker and len(row) > 2:
                speaker = row[1]
            else:
                speaker = None

            wav_path: Optional[Path] = None
            for wav_dir in wav_dirs:
                potential_paths: List[Path] = [
                    wav_dir / filename,
                    wav_dir / f"{filename}.wav",
                    wav_dir / f"{filename.lstrip('0')}.wav"
                ]
                for path in potential_paths:
                    if path.exists():
                        wav_path = path
                        break
                if wav_path:
                    break

            if not config.skip_audio and not wav_path:
                _LOGGER.warning("Missing audio file for filename: %s", filename)
                continue

            if not config.skip_audio and wav_path and wav_path.stat().st_size == 0:
                _LOGGER.warning("Empty audio file: %s", wav_path)
                continue

            # Ensure wav_path is Path or None, and is never accessed if skip_audio is true
            yield Utterance(
                text=text,
                audio_path=wav_path or Path(""), # Use empty path if skipping audio, should not be used
                speaker=speaker,
                speaker_id=config.speaker_id,
            )


def phonemize_worker(
        config: PreprocessorConfig,
        task_queue: JoinableQueue,
        result_queue: Queue,
        phonemizer: Phonemizer,
) -> None:
    """
    Worker process for phonemization and audio processing.

    Args:
        config: The configuration object containing runtime parameters.
        task_queue: Queue for receiving batches of Utterance objects.
        result_queue: Queue for sending processed results (Utterance, set of phonemes).
        phonemizer: The initialized Phonemizer instance.
    """
    try:
        casing: Callable[[str], str] = get_text_casing(config.text_casing)
        silence_detector = make_silence_detector()

        while True:
            # Get a batch of utterances to process
            utterance_batch: Union[List[Utterance], None] = task_queue.get()
            if utterance_batch is None:
                # Signal to exit
                task_queue.task_done()
                break

            for utt in utterance_batch:
                try:
                    # Normalize text (case, numbers, etc.)
                    utterance: str = casing(normalize(utt.text, config.language))

                    # Add diacritics
                    if config.add_diacritics:
                        utterance = phonemizer.add_diacritics(utterance, config.language)

                    # Phonemize the text
                    utt.phonemes = [p for p in phonemizer.phonemize_to_list(utterance, config.language)
                                    if p != "\n"] # HACK: not sure where this is coming from
                    if not utt.phonemes:
                        raise RuntimeError(f"Phonemes not found for '{utterance}'")

                    # Process audio if not skipping
                    if not config.skip_audio:
                        utt.audio_norm_path, utt.audio_spec_path = cache_norm_audio(
                            utt.audio_path,
                            config.cache_dir,
                            silence_detector,
                            config.sample_rate,
                        )

                    # Put the processed utterance and its phonemes into the result queue
                    # The result is a tuple of (Utterance, set of unique phonemes in that utterance)
                    result_queue.put((utt, set(utt.phonemes)))
                except Exception:
                    _LOGGER.exception("Failed to process utterance: %s", utt.audio_path)
                    result_queue.put((None, set()))

            task_queue.task_done()

    except Exception:
        _LOGGER.exception("Worker process failed")


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "-i",
    "--input-dir",
    "input_dir",
    type=click.Path(exists=True, file_okay=False, path_type=Path),
    required=True,
    help="Directory with audio dataset (e.g., containing metadata.csv and wavs/)",
)
@click.option(
    "-o",
    "--output-dir",
    "output_dir",
    type=click.Path(file_okay=False, path_type=Path),
    required=True,
    help="Directory to write output files for training (config.json, dataset.jsonl)",
)
@click.option(
    "-l",
    "--language",
    "language",
    required=True,
    help="phonemizer language code (e.g., 'en', 'es', 'fr')",
)
@click.option(
    "-c",
    "--prev-config",
    "prev_config",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    default=None,
    help="Optional path to a previous config.json from which to reuse phoneme_id_map. (for fine-tuning only)",
)
@click.option(
    "--drop-extra-phonemes",
    "drop_extra_phonemes",
    type=bool,
    default=True,
    help="If training data has more symbols than base model, discard new symbols. (for fine-tuning only)",
)
@click.option(
    "-r",
    "--sample-rate",
    "sample_rate",
    type=int,
    default=22050,
    help="Target sample rate for voice (hertz, Default: 22050)",
)
@click.option(
    "--cache-dir",
    "cache_dir",
    type=click.Path(file_okay=False, path_type=Path),
    default=None,
    help="Directory to cache processed audio files. Defaults to <output-dir>/cache/<sample-rate>.",
)
@click.option(
    "-w",
    "--max-workers",
    "max_workers",
    type=click.IntRange(min=1),
    default=os.cpu_count() or 1,
    help="Maximum number of worker processes to use for parallel processing. Defaults to CPU count.",
)
@click.option(
    "--single-speaker",
    "single_speaker",
    is_flag=True,
    help="Force treating the dataset as single speaker, ignoring metadata speaker columns.",
)
@click.option(
    "--speaker-id",
    "speaker_id",
    type=int,
    default=None,
    help="Specify a fixed speaker ID (0, 1, etc.) for a single speaker dataset.",
)
@click.option(
    "--phoneme-type",
    "phoneme_type",
    type=click.Choice([p.value for p in PhonemeType]),
    default=PhonemeType.ESPEAK.value,
    help="Type of phonemes to use.",
)
@click.option(
    "--alphabet",
    "alphabet",
    type=click.Choice([a.value for a in Alphabet]),
    default=Alphabet.IPA.value,
    help="Phoneme alphabet to use (e.g., IPA).",
)
@click.option(
    "--phonemizer-model",
    "phonemizer_model",
    default="",
    help="Path or name of a custom phonemizer model, if applicable.",
)
@click.option(
    "--text-casing",
    "text_casing",
    type=click.Choice(("ignore", "lower", "upper", "casefold")),
    default="ignore",
    help="Casing applied to utterance text before phonemization.",
)
@click.option(
    "--dataset-name",
    "dataset_name",
    default=None,
    help="Name of dataset to put in config (default: name of <output_dir>/../).",
)
@click.option(
    "--audio-quality",
    "audio_quality",
    default=None,
    help="Audio quality description to put in config (default: name of <output_dir>).",
)
@click.option(
    "--skip-audio",
    "skip_audio",
    is_flag=True,
    help="Do not preprocess or cache audio files.",
)
@click.option(
    "--debug",
    "debug",
    is_flag=True,
    help="Print DEBUG messages to the console.",
)
@click.option(
    "--add-diacritics",
    "add_diacritics",
    is_flag=True,
    help="Add diacritics to text (phonemizer specific, e.g., to denote stress).",
)
@click.option(
    "--jsonl-audio-path",
    default=None,
    help="override audio_path base directory (everything before '/wav') in generated dataset.jsonl"
)
@click.option(
    "--jsonl-audio-spec-path",
    default=None,
    help="override audio_norm_path/audio_spec_path base directory (everything before '/cache') in generated dataset.jsonl"
)
def cli(
    input_dir: Path,
    output_dir: Path,
    language: str,
    prev_config: Path,
    drop_extra_phonemes: bool,
    sample_rate: int,
    cache_dir: Optional[Path],
    max_workers: Optional[int],
    single_speaker: bool,
    speaker_id: Optional[int],
    phoneme_type: str,
    alphabet: str,
    phonemizer_model: str,
    text_casing: str,
    dataset_name: Optional[str],
    audio_quality: Optional[str],
    skip_audio: bool,
    debug: bool,
    add_diacritics: bool,
    jsonl_audio_path: Optional[str],
    jsonl_audio_spec_path: Optional[str],
) -> None:
    """
    Preprocess a TTS dataset (e.g., LJSpeech format) for training a VITS-style model.
    This script handles text normalization, phonemization, and optional audio caching.
    """
    # Create a config object from click arguments for easier passing
    config = PreprocessorConfig(
        input_dir=input_dir,
        output_dir=output_dir,
        language=language,
        sample_rate=sample_rate,
        cache_dir=cache_dir or output_dir / "cache" / str(sample_rate),
        max_workers=max_workers or os.cpu_count() or 1,
        single_speaker=single_speaker,
        speaker_id=speaker_id,
        phoneme_type=PhonemeType(phoneme_type),
        alphabet=Alphabet(alphabet),
        phonemizer_model=phonemizer_model,
        text_casing=text_casing,
        dataset_name=dataset_name,
        audio_quality=audio_quality,
        skip_audio=skip_audio,
        debug=debug,
        add_diacritics=add_diacritics,
    )

    # Setup logging
    level = logging.DEBUG if config.debug else logging.INFO
    logging.basicConfig(level=level)
    logging.getLogger().setLevel(level)
    logging.getLogger("numba").setLevel(logging.WARNING)

    # Validation
    if config.single_speaker and (config.speaker_id is not None):
        _LOGGER.fatal("--single-speaker and --speaker-id cannot both be provided")
        raise click.Abort()

    # Create directories
    config.output_dir.mkdir(parents=True, exist_ok=True)
    config.cache_dir.mkdir(parents=True, exist_ok=True)

    # Load all utterances from the dataset
    _LOGGER.info("Loading utterances from dataset...")
    utterances: List[Utterance] = list(ljspeech_dataset(config))
    if not utterances:
        _LOGGER.error("No valid utterances found in dataset.")
        return

    num_utterances: int = len(utterances)
    _LOGGER.info("Found %d utterances.", num_utterances)

    # Count speakers and assign IDs
    speaker_counts: Counter[str] = Counter(u.speaker for u in utterances if u.speaker)
    is_multispeaker: bool = len(speaker_counts) > 1
    speaker_ids: Dict[str, int] = {}
    if is_multispeaker:
        _LOGGER.info("%s speakers detected", len(speaker_counts))
        # Assign speaker ids by most number of utterances first
        for speaker_id, (speaker, _) in enumerate(speaker_counts.most_common()):
            speaker_ids[speaker] = speaker_id
    else:
        _LOGGER.info("Single speaker dataset")

    # --- Single Pass: Process audio/phonemes and collect results ---
    _LOGGER.info("Starting single pass processing with %d workers...", config.max_workers)

    # Initialize the phonemizer only once in the main process
    phonemizer: Phonemizer = get_phonemizer(config.phoneme_type,
                                            config.alphabet,
                                            config.phonemizer_model)

    batch_size: int = max(1, int(num_utterances / (config.max_workers * 2)))

    task_queue: "JoinableQueue[Optional[List[Utterance]]]" = JoinableQueue()
    # The result queue will hold tuples of (Utterance, set(phonemes))
    result_queue: "Queue[Tuple[Optional[Utterance], Set[str]]]" = Queue()

    # Start workers
    processes: List[Process] = [
        Process(
            target=phonemize_worker,
            args=(config, task_queue, result_queue, phonemizer)
        )
        for _ in range(config.max_workers)
    ]

    for proc in processes:
        proc.start()

    # Populate the task queue with batches
    task_count: int = 0
    for utt_batch in batched(utterances, batch_size):
        task_queue.put(utt_batch)
        task_count += len(utt_batch)

    # Signal workers to stop
    for _ in range(config.max_workers):
        task_queue.put(None)

    # Collect results from the queue with a progress bar
    processed_utterances: List[Utterance] = []
    all_phonemes: Set[str] = set()
    for _ in tqdm(range(task_count), desc="Processing utterances"):
        result: Tuple[Optional[Utterance], Set[str]] = result_queue.get()
        utt, unique_phonemes = result
        if utt is not None:
            processed_utterances.append(utt)
            all_phonemes.update(unique_phonemes)

    # Wait for workers to finish
    task_queue.join()
    for proc in processes:
        proc.join()


    # --- Build the final phoneme map from the collected phonemes ---
    _LOGGER.info("Building a phoneme map from collected dataset phonemes...")

    if prev_config:
        with open(prev_config) as f:
            cfg = json.load(f)
        # flatten list, same models (eg. piper) use a list of ids
        prev_phoneme_id_map = {k: v if not isinstance(v, list) else v[0]
                               for k, v in cfg["phoneme_id_map"].items()}

        prev_num_symbols = cfg.get("num_symbols", MAX_PHONEMES)
        _LOGGER.info(f"Loaded phoneme map from previous config: '{prev_config}'")
        all_phonemes.update(prev_phoneme_id_map.keys())
        final_phoneme_id_map = prev_phoneme_id_map
        _LOGGER.info("previous phoneme map contains %d phonemes.", len(final_phoneme_id_map))
    else:
        prev_num_symbols = MAX_PHONEMES
        final_phoneme_id_map: Dict[str, int] = DEFAULT_SPECIAL_PHONEME_ID_MAP.copy()
        if phonemizer.alphabet == Alphabet.IPA:
            all_phonemes.update(DEFAULT_IPA_PHONEME_ID_MAP.keys())

    # Filter out tokens that are already in the map
    existing_keys: Set[str] = set(final_phoneme_id_map.keys())
    new_phonemes: List[str] = sorted([p for p in all_phonemes
                                      if p not in existing_keys]
                                     )

    _LOGGER.info("Collected %d new phonemes.", len(new_phonemes))

    finetune_error = prev_config and len(new_phonemes)
    if finetune_error:
        if not drop_extra_phonemes:
            raise ValueError("training data contains different phonemes than previous phoneme map! Can not finetune model")
        else:
            _LOGGER.error("training data contains different phonemes than previous phoneme map! "
                          "Discarding new phonemes to still allow model finetuning")

    current_id: int = len(final_phoneme_id_map)
    for pho in new_phonemes:
        if finetune_error:
            _LOGGER.info(f"Discarded phoneme: {pho}")
        else:
            final_phoneme_id_map[pho] = current_id
            current_id += 1
            _LOGGER.debug(f"New phoneme: {pho}")

    if new_phonemes:
        _LOGGER.info("Final phoneme map contains %d phonemes.", len(final_phoneme_id_map))

    # --- Write the final config.json ---
    _LOGGER.info("Writing dataset config...")
    audio_quality = config.audio_quality or config.output_dir.name
    dataset_name = config.dataset_name or config.output_dir.parent.name

    config_data: Dict[str, Any] = {
        "dataset": dataset_name,
        "audio": {
            "sample_rate": config.sample_rate,
            "quality": audio_quality,
        },
        "lang_code": config.language,
        "inference": {"noise_scale": 0.667,
                      "length_scale": 1,
                      "noise_w": 0.8,
                      "add_diacritics": config.add_diacritics},
        "alphabet": phonemizer.alphabet.value,
        "phoneme_type": config.phoneme_type.value,
        "phonemizer_model": config.phonemizer_model,
        "phoneme_id_map": final_phoneme_id_map,
        "num_symbols": prev_num_symbols if prev_config else len(final_phoneme_id_map),
        "num_speakers": len(speaker_counts) if is_multispeaker else 1,
        "speaker_id_map": speaker_ids,
        "phoonnx_version": VERSION_STR,
    }

    with open(config.output_dir / "config.json", "w", encoding="utf-8") as config_file:
        json.dump(config_data, config_file, ensure_ascii=False, indent=2)

    # --- Apply final phoneme IDs and write dataset.jsonl ---
    _LOGGER.info("Writing dataset.jsonl...")
    valid_utterances_count: int = 0
    with open(config.output_dir / "dataset.jsonl", "w", encoding="utf-8") as dataset_file:
        for utt in processed_utterances:
            if is_multispeaker and utt.speaker is not None:
                if utt.speaker not in speaker_ids:
                    _LOGGER.error("Speaker '%s' not in speaker_id_map. This indicates an issue with your metadata.csv file.", utt.speaker)
                    continue
                utt.speaker_id = speaker_ids[utt.speaker]

            # Apply the final phoneme ID map to each utterance
            if utt.phonemes:
                utt.phoneme_ids = phonemes_to_ids(utt.phonemes, id_map=final_phoneme_id_map)

            if not utt.phoneme_ids:
                _LOGGER.warning("Skipping utterance with invalid phoneme_ids before writing: %s", utt.audio_path)
                continue

            # apply path overrides if needed
            # this allows pre-processing the dataset in one system and then train in other
            if jsonl_audio_path:
                base_path, fname = str(utt.audio_path).split("/wav/")
                utt.audio_path = Path(f"{jsonl_audio_path}/wav/{fname}")
            if jsonl_audio_spec_path:
                base_path, fname = str(utt.audio_norm_path).split("/cache/")
                utt.audio_norm_path = Path(f"{jsonl_audio_spec_path}/cache/{fname}")
                base_path, fname = str(utt.audio_spec_path).split("/cache/")
                utt.audio_spec_path = Path(f"{jsonl_audio_spec_path}/cache/{fname}")

            json.dump(
                utt.asdict(),
                dataset_file,
                ensure_ascii=False,
                cls=PathEncoder,
            )
            print("", file=dataset_file)
            valid_utterances_count += 1

    _LOGGER.info("Preprocessing complete. Wrote %d valid utterances to dataset.jsonl.", valid_utterances_count)


# -----------------------------------------------------------------------------

def batched(iterable: Iterable[Any], n: int) -> Iterable[List[Any]]:
    """
    Batch data from an iterable into lists of length n. The last batch may be shorter.

    Args:
        iterable: The input iterable to be batched.
        n: The desired size of each batch.

    Yields:
        List[Any]: A list representing a batch of items.
    """
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    batch = list(itertools.islice(it, n))
    while batch:
        yield batch
        batch = list(itertools.islice(it, n))


if __name__ == "__main__":
    cli()



from phoonnx.config import Alphabet
from phoonnx.phonemizers.base import BasePhonemizer
from phoonnx.thirdparty.bw2ipa import translate as bw2ipa
from phoonnx.thirdparty.mantoq import g2p as mantoq


class MantoqPhonemizer(BasePhonemizer):

    def __init__(self, alphabet=Alphabet.BUCKWALTER):
        if alphabet not in [Alphabet.IPA, Alphabet.BUCKWALTER]:
            raise ValueError("unsupported alphabet")
        super().__init__(alphabet)

    @classmethod
    def get_lang(cls, target_lang: str) -> str:
        """
        Validates and returns the closest supported language code.

        Args:
            target_lang (str): The language code to validate.

        Returns:
            str: The validated language code.

        Raises:
            ValueError: If the language code is unsupported.
        """
        # this check is here only to throw an exception if invalid language is provided
        return cls.match_lang(target_lang, ["ar"])

    def phonemize_string(self, text: str, lang: str = "ar") -> str:
        """
        Phonemizes an Arabic string using the Mantoq G2P tool.
        If the alphabet is set to IPA, it then converts the result using bw2ipa.
        """
        lang = self.get_lang(lang)
        # The mantoq function returns a tuple of (normalized_text, phonemes)
        normalized_text, phonemes = mantoq(text)

        # The phonemes are a list of characters, we join them into a string
        # and replace the word separator token with a space.
        phonemes = "".join(phonemes).replace("_+_", " ")

        if self.alphabet == Alphabet.IPA:
            # If the alphabet is IPA, we use the bw2ipa function to translate
            # the Buckwalter-like phonemes into IPA.
            return bw2ipa(phonemes)

        # Otherwise, we return the phonemes in the default Mantoq alphabet.
        return phonemes


if __name__ == "__main__":
    from phoonnx.phonemizers.mul import EspeakPhonemizer

    espeak = EspeakPhonemizer()

    # Initialize phonemizers for both MANTOQ and IPA alphabets
    pho_mantoq = MantoqPhonemizer(alphabet=Alphabet.IPA)


    def compare(text):
        print(f"Original Text: {text}")
        print(f"  Mantoq: {pho_mantoq.phonemize_string(text, 'ar')}")
        print(f"  Espeak: {espeak.phonemize_string(text, 'ar')}")

        ts = pho_mantoq.add_diacritics(text, 'ar')
        print(f"Tashkeel Text: {ts}")
        print(f"  Mantoq: {pho_mantoq.phonemize_string(ts, 'ar')}")
        print(f"  Espeak: {espeak.phonemize_string(ts, 'ar')}")
        print("\n#########################")


    text = "مرحبا بالعالم"
    compare(text)

    text = "ذهب الطالب إلى المكتبة لقراءة كتاب عن تاريخ الأندلس."
    compare(text)

    # 1. Test for gemination of a sun letter (e.g., ash-shams)
    text = "الشمس"
    compare(text)

    # 2. Test for long vowels (e.g., 'fil' - elephant)
    text = "فيل"
    compare(text)

    # 3. Test for glide (e.g., 'yawm' - day)
    text = "يوم"
    compare(text)

    # 4. Test for long vowels (e.g., 'suwr' - wall)
    text = "سور"
    compare(text)

    # 5. Test for glide (e.g., 'law' - if)
    text = "لو"
    compare(text)



from phoonnx.phonemizers.base import BasePhonemizer, Alphabet


class MirandesePhonemizer(BasePhonemizer):
    _LANGS = ["mwl"]

    def __init__(self):
        super().__init__(Alphabet.IPA)
        from mwl_phonemizer import CRFOrthoCorrector
        self.pho = CRFOrthoCorrector()

    @classmethod
    def get_lang(cls, target_lang: str) -> str:
        """
        Validates and returns the closest supported language code.

        Args:
            target_lang (str): The language code to validate.

        Returns:
            str: The validated language code.

        Raises:
            ValueError: If the language code is unsupported.
        """
        return cls.match_lang(target_lang, cls._LANGS)

    def phonemize_string(self, text: str, lang: str) -> str:
        # Validate language is supported
        lang = self.get_lang(lang)
        return self.pho.phonemize_sentence(text)


if __name__ == "__main__":
    pho = MirandesePhonemizer()
    print(pho.phonemize_string("ls", "mwl"))


import os

from setuptools import setup

BASEDIR = os.path.abspath(os.path.dirname(__file__))


def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


def required(requirements_file):
    """ Read requirements file and remove comments and empty lines. """
    with open(os.path.join(os.path.dirname(__file__), requirements_file),
              'r') as f:
        requirements = f.read().splitlines()
        return [pkg for pkg in requirements
                if pkg.strip() and not pkg.startswith("#")]


def get_version():
    """ Find the version of the package"""
    version_file = os.path.join(BASEDIR, 'phoonnx', 'version.py')
    major, minor, build, alpha = (None, None, None, None)
    with open(version_file) as f:
        for line in f:
            if 'VERSION_MAJOR' in line:
                major = line.split('=')[1].strip()
            elif 'VERSION_MINOR' in line:
                minor = line.split('=')[1].strip()
            elif 'VERSION_BUILD' in line:
                build = line.split('=')[1].strip()
            elif 'VERSION_ALPHA' in line:
                alpha = line.split('=')[1].strip()

            if ((major and minor and build and alpha) or
                    '# END_VERSION_BLOCK' in line):
                break
    version = f"{major}.{minor}.{build}"
    if alpha and int(alpha) > 0:
        version += f"a{alpha}"
    return version


extra_files = package_files('phoonnx')

PLUGIN_ENTRY_POINT = 'ovos-tts-plugin-phoonnx = phoonnx.opm:PhoonnxTTSPlugin'

setup(
    name='phoonnx',
    version=get_version(),
    packages=['phoonnx',
              'phoonnx.phonemizers',
              'phoonnx.thirdparty',
              'phoonnx.thirdparty.kog2p',
              'phoonnx.thirdparty.tashkeel',
              'phoonnx.thirdparty.mantoq',
              'phoonnx.thirdparty.mantoq.buck',
              'phoonnx.thirdparty.mantoq.pyarabic',
              'phoonnx_train',
              'phoonnx_train.vits',
              'phoonnx_train.vits.monotonic_align',
              'phoonnx_train.norm_audio',
              'phoonnx_train.norm_audio.models'],
    include_package_data=True,
    package_data={'': extra_files},
    install_requires=required('requirements/requirements.txt'),
    extras_require={
        x.split(".")[0]: required(f'requirements/{x}')
        for x in os.listdir(f"{BASEDIR}/requirements")
        if x.endswith(".txt") and x != "requirements.txt"

    },
    url='https://github.com/TigreGotico/phoonnx',
    license='',
    author='JarbasAi',
    author_email='jarbasai@mailfence.com',
    description='',
    entry_points={
        'mycroft.plugin.tts': PLUGIN_ENTRY_POINT,
        'console_scripts': [
            'phoonnx-voices=phoonnx.cli:cli'
        ]
    }
)
